"""
Telnet server for x84.

Limitations:

- No linemode support, character-at-a-time only.
- No out-of-band / data mark (DM) / sync supported
- No flow control (``^S``, ``^Q``)

This is a modified version of miniboa retrieved from svn address
http://miniboa.googlecode.com/svn/trunk/miniboa which is meant for
MUD's. This server would not be safe for most (linemode) MUD clients.

Changes from miniboa:

- character-at-a-time input instead of linemode
- encoding option on send
- strict rejection of linemode
- terminal type detection
- environment variable support
- GA and SGA
- utf-8 safe
"""
# ------------------------------------------------------------------------------
#   miniboa/async.py
#   miniboa/telnet.py
#
#   Copyright 2009 Jim Storch
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain a
#   copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
# ------------------------------------------------------------------------------

from __future__ import absolute_import

# std
import socket
import array
import time
import logging
import select
import errno
from telnetlib import LINEMODE, NAWS, NEW_ENVIRON, ENCRYPT, AUTHENTICATION
from telnetlib import BINARY, SGA, ECHO, STATUS, TTYPE, TSPEED, LFLOW
from telnetlib import XDISPLOC, IAC, DONT, DO, WONT, WILL, SE, NOP, DM, BRK
from telnetlib import IP, AO, AYT, EC, EL, GA, SB

# local
from x84.bbs.exception import Disconnected
from .terminal import spawn_client_session, on_naws
from .client import BaseClient, BaseConnect
from .server import BaseServer

IS = chr(0)  # Sub-process negotiation IS command
SEND = chr(1)  # Sub-process negotiation SEND command
UNSUPPORTED_WILL = (LINEMODE, LFLOW, TSPEED, ENCRYPT, AUTHENTICATION)

#---[ Telnet Notes ]-----------------------------------------------------------
# (See RFC 854 for more information)
#
# Negotiating a Local Option
# --------------------------
#
# Side A begins with:
#
#    "IAC WILL/WONT XX"   Meaning "I would like to [use|not use] option XX."
#
# Side B replies with either:
#
#    "IAC DO XX"     Meaning "OK, you may use option XX."
#    "IAC DONT XX"   Meaning "No, you cannot use option XX."
#
#
# Negotiating a Remote Option
# ----------------------------
#
# Side A begins with:
#
#    "IAC DO/DONT XX"  Meaning "I would like YOU to [use|not use] option XX."
#
# Side B replies with either:
#
#    "IAC WILL XX"   Meaning "I will begin using option XX"
#    "IAC WONT XX"   Meaning "I will not begin using option XX"
#
#
# The syntax is designed so that if both parties receive simultaneous requests
# for the same option, each will see the other's request as a positive
# acknowledgment of it's own.
#
# If a party receives a request to enter a mode that it is already in, the
# request should not be acknowledged.

# Where you see DE in my comments I mean 'Distant End', e.g. the client.

UNKNOWN = -1

#-----------------------------------------------------------------Telnet Option


class TelnetOption(object):

    """
    Simple class used to track the status of an extended Telnet option.

    Attributes and their state values:

    - ``local_option``: UNKNOWN (default), True, or False.
    - ``remote_option``: UNKNOWN (default), True, or False.
    - ``reply_pending``: True or Fale.
    """
    # pylint: disable=R0903
    #         Too few public methods (0/2)

    def __init__(self):
        """
        Set attribute defaults on init.
        """
        self.local_option = UNKNOWN     # Local state of an option
        self.remote_option = UNKNOWN    # Remote state of an option
        self.reply_pending = False      # Are we expecting a reply?


def name_option(option):
    """
    Perform introspection of global CONSTANTS for equivalent values,
    and return a string that displays its possible meanings
    """
    values = ';?'.join([k for k, v in globals().iteritems()
                        if option == v and k not in ('SEND', 'IS',)])
    return values if values != '' else str(ord(option))


def debug_option(func):
    """
    This function is a decorator that debug prints the 'from' address for
    callables decorated with this. This helps during telnet negotiation, to
    understand which function sets or checks local or remote option states.
    """
    import os
    import inspect

    def wrapper(self, *args):
        """
        inner wrapper for debug_option
        """
        stack = inspect.stack()
        self.log.debug('%s:%s %s(%s%s)',
                       os.path.basename(stack[1][1]), stack[1][2],
                       func.__name__, name_option(args[0]),
                       ', %s' % (args[1],) if len(args) == 2 else '')
        return func(self, *args)
    return wrapper


#------------------------------------------------------------------------Telnet

class TelnetClient(BaseClient):

    """
    Represents a remote Telnet Client, instantiated from TelnetServer.
    """
    # pylint: disable=R0902,R0904
    #         Too many instance attributes
    #         Too many public methods

    kind = 'telnet'

    #: maximum size of telnet subnegotiation string, allowing for a fairly
    #: large value for NEW_ENVIRON.
    SB_MAXLEN = 65534

    def __init__(self, sock, address_pair, on_naws=None):
        super(TelnetClient, self).__init__(sock, address_pair, on_naws)
        self.telnet_sb_buffer = array.array('c')

        # State variables for interpreting incoming telnet commands
        self.telnet_got_iac = False
        self.telnet_got_cmd = None
        self.telnet_got_sb = False
        self.telnet_opt_dict = {}

        self.ENV_REQUESTED = False
        self.ENV_REPLIED = False

    def request_will_sga(self):
        """
        Request DE to Suppress Go-Ahead.  See RFC 858.
        """
        self._iac_will(SGA)
        self._note_reply_pending(SGA, True)

    def request_will_echo(self):
        """
        Tell the DE that we would like to echo their text.  See RFC 857.
        """
        self._iac_will(ECHO)
        self._note_reply_pending(ECHO, True)

    def request_will_binary(self):
        """
        Tell the DE that we would like to use binary 8-bit (utf8).
        """
        self._iac_will(BINARY)
        self._note_reply_pending(BINARY, True)

    def request_do_binary(self):
        """
        Tell the DE that we would like them to input binary 8-bit (utf8).
        """
        self._iac_do(BINARY)
        self._note_reply_pending(BINARY, True)

    def request_do_sga(self):
        """
        Request to Negotiate SGA.  See ...
        """
        self._iac_do(SGA)
        self._note_reply_pending(SGA, True)

    def request_do_naws(self):
        """
        Request to Negotiate About Window Size.  See RFC 1073.
        """
        self._iac_do(NAWS)
        self._note_reply_pending(NAWS, True)

    def request_do_env(self):
        """
        Request to Negotiate About Window Size.  See RFC 1073.
        """
        self._iac_do(NEW_ENVIRON)
        self._note_reply_pending(NEW_ENVIRON, True)

    def request_env(self):
        """
        Request sub-negotiation NEW_ENVIRON. See RFC 1572.
        """
        if self.ENV_REQUESTED:
            return  # avoid asking twice ..
        rstr = bytes(''.join((IAC, SB, NEW_ENVIRON, SEND, chr(0))))
        rstr += bytes(chr(0).join(
            ("USER TERM SHELL COLUMNS LINES C_CTYPE XTERM_LOCALE DISPLAY "
             "SSH_CLIENT SSH_CONNECTION SSH_TTY HOME HOSTNAME PWD MAIL LANG "
             "PWD UID USER_ID EDITOR LOGNAME".split())))
        rstr += bytes(''.join((chr(3), IAC, SE)))
        self.ENV_REQUESTED = True
        self.send_str(rstr)

    def request_do_ttype(self):
        """
        Begins TERMINAL-TYPE negotiation
        """
        if self.check_remote_option(TTYPE) in (False, UNKNOWN):
            self._iac_do(TTYPE)
            self._note_reply_pending(TTYPE, True)

    def request_ttype(self):
        """
        Sends IAC SB TTYPE SEND IAC SE
        """
        self.send_str(bytes(''.join((
            IAC, SB, TTYPE, SEND, IAC, SE))))

    def recv_ready(self):
        """
        Returns True if data is awaiting on the telnet socket.
        """
        return (self.is_active() and bool(
            select.select([self.sock.fileno()], [], [], 0)[0]))

    def socket_recv(self):
        """
        Called by TelnetServer.poll() when recv data is ready.  Read any
        data on socket, processing telnet commands, and buffering all
        other bytestrings to self.recv_buffer.  If data is not received,
        or the connection is closed, x84.bbs.exception.Disconnected is
        raised.
        """
        try:
            data = self.sock.recv(self.BLOCKSIZE_RECV)
            recv = len(data)
            if recv == 0:
                raise Disconnected('Closed by client (EOF)')

        except socket.error as err:
            if err.errno == errno.EWOULDBLOCK:
                return 0
            raise Disconnected('socket_recv error: {0}'.format(err))

        self.bytes_received += recv
        self.last_input_time = time.time()

        # Test for telnet commands, non-telnet bytes
        # are pushed to self.recv_buffer (side-effect),
        for byte in data:
            self._iac_sniffer(byte)
        return recv

    def send_unicode(self, ucs, encoding='utf8'):
        """ Buffer unicode string, encoded for client as 'encoding'. """
        # Must be escaped 255 (IAC + IAC) to avoid IAC interpretation.
        self.send_str(ucs.encode(encoding, 'replace').replace(IAC, 2 * IAC))

    def _recv_byte(self, byte):
        """
        Buffer non-telnet commands bytestrings into recv_buffer.
        """
        self.recv_buffer.fromstring(byte)

    def _iac_sniffer(self, byte):
        """
        Watches incomming data for Telnet IAC sequences.
        Passes the data, if any, with the IAC commands stripped to
        _recv_byte().
        """
        # Are we not currently in an IAC sequence coming from the DE?
        if self.telnet_got_iac is False:
            if byte == IAC:
                self.telnet_got_iac = True
            # Are we currently in a sub-negotiation?
            elif self.telnet_got_sb is True:
                self.telnet_sb_buffer.fromstring(byte)
                # Sanity check on length
                if len(self.telnet_sb_buffer) >= self.SB_MAXLEN:
                    raise Disconnected('sub-negotiation buffer filled')
            else:
                # Just a normal NVT character
                self._recv_byte(byte)
            return

        # Did we get sent a second IAC?
        if byte == IAC and self.telnet_got_sb is True:
            # Must be an escaped 255 (IAC + IAC)
            self.telnet_sb_buffer.fromstring(byte)
            self.telnet_got_iac = False
        # Do we already have an IAC + CMD?
        elif self.telnet_got_cmd is not None:
            # Yes, so handle the option
            self._three_byte_cmd(byte)
        # We have IAC but no CMD
        else:
            # Is this the middle byte of a three-byte command?
            if byte in (DO, DONT, WILL, WONT):
                self.telnet_got_cmd = byte
            else:
                # Nope, must be a two-byte command
                self._two_byte_cmd(byte)

    def _two_byte_cmd(self, cmd):
        """
        Handle incoming Telnet commands that are two bytes long.
        """
        # self.log.debug ('recv _two_byte_cmd %s', name_option(cmd),)
        if cmd == SB:
            # Begin capturing a sub-negotiation string
            self.telnet_got_sb = True
            self.telnet_sb_buffer = array.array('c')
        elif cmd == SE:
            # Stop capturing a sub-negotiation string
            self.telnet_got_sb = False
            self._sb_decoder()
        elif cmd == IAC:
            # IAC, IAC is used for a literal \xff character.
            self._recv_byte(IAC)
        elif cmd == IP:
            self.deactivate()
            self.log.info('{self.addrport} received (IAC, IP): closing.'
                          .format(self=self))
        elif cmd == AO:
            flushed = len(self.recv_buffer)
            self.recv_buffer = array.array('c')
            self.log.debug('Abort Output (AO); %s bytes discarded.', flushed)
        elif cmd == AYT:
            self.send_str(bytes('\b'))
            self.log.debug('Are You There (AYT); "\\b" sent.')
        elif cmd == EC:
            self.recv_buffer.fromstring('\b')
            self.log.debug('Erase Character (EC); "\\b" queued.')
        elif cmd == EL:
            self.log.warn('Erase Line (EC) received; ignored.')
        elif cmd == GA:
            self.log.warn('Go Ahead (GA) received; ignored.')
        elif cmd == NOP:
            self.log.debug('NUL ignored.')
        elif cmd == DM:
            self.log.warn('Data Mark (DM) received; ignored.')
        elif cmd == BRK:
            self.log.warn('Break (BRK) received; ignored.')
        else:
            self.log.error('_two_byte_cmd invalid: %r', cmd)
        self.telnet_got_iac = False
        self.telnet_got_cmd = None

    def _three_byte_cmd(self, option):
        """
        Handle incoming Telnet commmands that are three bytes long.
        """
        cmd = self.telnet_got_cmd
        self.log.debug('recv IAC %s %s', name_option(cmd), name_option(option))
        # Incoming DO's and DONT's refer to the status of this end

        if cmd == DO:
            self._handle_do(option)
        elif cmd == DONT:
            self._handle_dont(option)
        elif cmd == WILL:
            self._handle_will(option)
        elif cmd == WONT:
            self._handle_wont(option)
        else:
            self.log.debug('{self.addrport}: unhandled _three_byte_cmd: {opt}.'
                           .format(self=self, opt=name_option(option)))
        self.telnet_got_iac = False
        self.telnet_got_cmd = None

    def _handle_do(self, option):
        """
        Process a DO command option received by DE.
        """
        # pylint: disable=R0912
        #         TelnetClient._handle_do: Too many branches (13/12)
        # if any pending WILL options were send, they have been received
        self._note_reply_pending(option, False)
        if option == ECHO:
            # DE requests us to echo their input
            if self.check_local_option(ECHO) is not True:
                self._note_local_option(ECHO, True)
                self._iac_will(ECHO)
        elif option == BINARY:
            # DE requests to recv BINARY
            if self.check_local_option(BINARY) is not True:
                self._note_local_option(BINARY, True)
                self._iac_will(BINARY)
        elif option == SGA:
            # DE wants us to supress go-ahead
            if self.check_local_option(SGA) is not True:
                self._note_local_option(SGA, True)
                # always send DO SGA after WILL SGA, requesting the DE
                # also supress their go-ahead. this order seems to be the
                # 'magic sequence' to disable linemode on certain clients
                self._iac_will(SGA)
                self._iac_do(SGA)
        elif option == LINEMODE:
            # DE wants to do linemode editing
            # denied
            if self.check_local_option(option) is not False:
                self._note_local_option(option, False)
                self._iac_wont(LINEMODE)
        elif option == ENCRYPT:
            # DE is willing to receive encrypted data
            # denied
            if self.check_local_option(option) is not False:
                self._note_local_option(option, False)
                # let DE know we refuse to send encrypted data.
                self._iac_wont(ENCRYPT)
        elif option == STATUS:
            # DE wants to know if we support STATUS,
            if self.check_local_option(option) is not True:
                self._note_local_option(option, True)
                self._iac_will(STATUS)
                self._send_status()
        else:
            if self.check_local_option(option) is UNKNOWN:
                self._note_local_option(option, False)
                self.log.debug('{self.addrport}: unhandled do: {opt}.'
                               .format(self=self, opt=name_option(option)))
                self._iac_wont(option)

    def _send_status(self):
        """
        Process a DO STATUS sub-negotiation received by DE. (rfc859)
        """
        # warning:
        rstr = bytes(''.join((IAC, SB, STATUS, IS)))
        for opt, status in self.telnet_opt_dict.items():
            # my_want_state_is_will
            if status.local_option is True:
                self.log.debug('send WILL %s', name_option(opt))
                rstr += bytes(''.join((WILL, opt)))
            elif status.reply_pending is True and opt in (ECHO, SGA):
                self.log.debug('send WILL %s (want)', name_option(opt))
                rstr += bytes(''.join((WILL, opt)))
            # his_want_state_is_will
            elif status.remote_option is True:
                self.log.debug('send DO %s', name_option(opt))
                rstr += bytes(''.join((DO, opt)))
            elif (status.reply_pending is True
                    and opt in (NEW_ENVIRON, NAWS, TTYPE)):
                self.log.debug('send DO %s (want)', name_option(opt))
                rstr += bytes(''.join((DO, opt)))
        rstr += bytes(''.join((IAC, SE)))
        self.log.debug('send %s', ' '.join(name_option(opt) for opt in rstr))
        self.send_str(rstr)

    def _handle_dont(self, option):
        """
        Process a DONT command option received by DE.
        """
        self._note_reply_pending(option, False)
        if option == ECHO:
            # client demands we do not echo
            if self.check_local_option(ECHO) is not False:
                self._note_local_option(ECHO, False)
        elif option == BINARY:
            # client demands no binary mode
            if self.check_local_option(BINARY) is not False:
                self._note_local_option(BINARY, False)
        elif option == SGA:
            # DE demands that we start or continue transmitting
            # GAs (go-aheads) when transmitting data.
            if self.check_local_option(SGA) is not False:
                self._note_local_option(SGA, False)
        elif option == LINEMODE:
            # client demands no linemode.
            if self.check_remote_option(LINEMODE) is not False:
                self._note_remote_option(LINEMODE, False)
        else:
            self.log.debug('{self.addrport}: unhandled dont: {opt}.'
                           .format(self=self, opt=name_option(option)))

    def _handle_will(self, option):
        """
        Process a WILL command option received by DE.
        """
        # pylint: disable=R0912
        #        Too many branches (19/12)
        self._note_reply_pending(option, False)
        if option == ECHO:
            raise Disconnected(
                'Refuse WILL ECHO by client, closing connection.')
        elif option == BINARY:
            if self.check_remote_option(BINARY) is not True:
                self._note_remote_option(BINARY, True)
                # agree to use BINARY
                self._iac_do(BINARY)
        elif option == NAWS:
            if self.check_remote_option(NAWS) is not True:
                self._note_remote_option(NAWS, True)
                self._note_local_option(NAWS, True)
                # agree to use NAWS, / go ahead ?
                self._iac_do(NAWS)
        elif option == STATUS:
            if self.check_remote_option(STATUS) is not True:
                self._note_remote_option(STATUS, True)
                self.send_str(bytes(''.join((
                    IAC, SB, STATUS, SEND, IAC, SE))))  # go ahead
        elif option in UNSUPPORTED_WILL:
            if self.check_remote_option(option) is not False:
                # let DE know we refuse to do linemode, encryption, etc.
                self._iac_dont(option)
        elif option == SGA:
            #  IAC WILL SUPPRESS-GO-AHEAD
            #
            # The sender of this command requests permission to begin
            # suppressing transmission of the TELNET GO AHEAD (GA)
            # character when transmitting data characters, or the
            # sender of this command confirms it will now begin suppressing
            # transmission of GAs with transmitted data characters.
            if self.check_remote_option(SGA) is not True:
                # sender of this command confirms that the sender of data
                # is expected to suppress transmission of GAs.
                self._iac_do(SGA)
                self._note_remote_option(SGA, True)
        elif option == NEW_ENVIRON:
            if self.check_remote_option(NEW_ENVIRON) in (False, UNKNOWN):
                self._note_remote_option(NEW_ENVIRON, True)
                self.request_env()
            self._note_local_option(NEW_ENVIRON, True)
        elif option == XDISPLOC:
            # if they want to send it, go ahead.
            if self.check_remote_option(XDISPLOC):
                self._note_remote_option(XDISPLOC, True)
                self._iac_do(XDISPLOC)
                self.send_str(bytes(''.join((
                    IAC, SB, XDISPLOC, SEND, IAC, SE))))
        elif option == TTYPE:
            if self.check_remote_option(TTYPE) in (False, UNKNOWN):
                self._note_remote_option(TTYPE, True)
                self.request_ttype()
        else:
            self.log.debug('{self.addrport}: unhandled will: {opt} (ignored).'
                           .format(self=self, opt=name_option(option)))

    def _handle_wont(self, option):
        """
        Process a WONT command option received by DE.
        """
        # pylint: disable=R0912
        #         TelnetClient._handle_wont: Too many branches (13/12)
        self._note_reply_pending(option, False)
        if option == ECHO:
            if self.check_remote_option(ECHO) in (True, UNKNOWN):
                self._note_remote_option(ECHO, False)
                self._iac_dont(ECHO)
        elif option == BINARY:
            # client demands no binary mode
            if self.check_remote_option(BINARY) in (True, UNKNOWN):
                self._note_remote_option(BINARY, False)
                self._iac_dont(BINARY)
        elif option == SGA:
            if self._check_reply_pending(SGA):
                self._note_reply_pending(SGA, False)
                self._note_remote_option(SGA, False)
            elif self.check_remote_option(SGA) in (True, UNKNOWN):
                self._note_remote_option(SGA, False)
                self._iac_dont(SGA)
        elif option == TTYPE:
            if self._check_reply_pending(TTYPE):
                self._note_reply_pending(TTYPE, False)
                self._note_remote_option(TTYPE, False)
            elif self.check_remote_option(TTYPE) in (True, UNKNOWN):
                self._note_remote_option(TTYPE, False)
                self._iac_dont(TTYPE)
        elif option in (NEW_ENVIRON, NAWS):
            if self._check_reply_pending(option):
                self._note_reply_pending(option, False)
                self._note_remote_option(option, False)
            elif self.check_remote_option(option) in (True, UNKNOWN):
                self._note_remote_option(option, False)
        else:
            self.log.debug('{self.addrport}: unhandled wont: {opt}.'
                           .format(self=self, opt=name_option(option)))
            self._note_remote_option(option, False)

    def _sb_decoder(self):
        """
        Figures out what to do with a received sub-negotiation block.
        """

        buf = self.telnet_sb_buffer
        if 0 == len(buf):
            self.log.error('nil SB')
            return
        self.log.debug('recv SB: %s %s',
                       name_option(buf[0]),
                       'IS %r' % (buf[2:],) if len(buf) > 1 and buf[1] is IS
                       else repr(buf[1:]))
        if 1 == len(buf) and buf[0] == chr(0):
            self.log.error('0nil SB')
            return
        elif len(buf) < 2:
            self.log.error('SB too short')
            return
        elif (TTYPE, IS) == (buf[0], buf[1]):
            self._sb_ttype(buf[2:].tostring())
        elif (XDISPLOC, IS) == (buf[0], buf[1]):
            self._sb_xdisploc(buf[2:].tostring())
        elif (NEW_ENVIRON, IS) == (buf[0], buf[1]):
            self._sb_env(buf[2:].tostring())
        elif NAWS == buf[0]:
            self._sb_naws(buf)
        elif (STATUS, SEND) == (buf[0], buf[1]):
            self._send_status()
        else:
            self.log.error('unsupported subnegotiation, %s: %r',
                           name_option(buf[0]), buf,)
        self.telnet_sb_buffer = ''

    def _sb_xdisploc(self, bytestring):
        """
        Process incoming sub-negotiation XDISPLOC
        """
        prev_display = self.env.get('DISPLAY', None)
        if prev_display is None:
            self.log.info("env['DISPLAY'] = %r.", bytestring)
        elif prev_display != bytestring:
            self.log.info("env['DISPLAY'] = %r by XDISPLOC was:%s.",
                          bytestring, prev_display)
        else:
            self.log.debug('XDSIPLOC ignored (DISPLAY already set).')
        self.env['DISPLAY'] = bytestring

    def _sb_ttype(self, bytestring):
        """
        Processes incoming subnegotiation TTYPE
        """
        term_str = bytestring.lower().strip()
        while term_str.endswith('\x00'):
            term_str = term_str[:-1]  # netrunner did this ..
        prev_term = self.env.get('TERM', None)
        if prev_term is None:
            self.log.debug("env['TERM'] = %r.", term_str)
        elif prev_term != term_str:
            self.log.debug("env['TERM'] = %r by TTYPE%s.", term_str,
                           ', was: %s' % (prev_term,)
                           if prev_term != self.TTYPE_UNDETECTED else '')
        else:
            self.log.debug('TTYPE ignored (TERM already set).')
        self.env['TERM'] = term_str

    def _sb_env(self, bytestring):
        """
        Processes incoming sub-negotiation NEW_ENVIRON
        """
        breaks = list([idx for (idx, byte) in enumerate(bytestring)
                       if byte in (chr(0), chr(3))])
        for start, end in zip(breaks, breaks[1:]):
            pair = bytestring[start + 1:end].split(chr(1))
            if len(pair) == 1:
                if (pair[0] in self.env
                        and pair[0] not in ('LINES', 'COLUMNS', 'TERM')):
                    self.log.warn("del env[%r]", pair[0])
                    del self.env[pair[0]]
            elif len(pair) == 2:
                if pair[0] == 'TERM':
                    pair[1] = pair[1].lower()
                overwrite = (pair[0] == 'TERM'
                             and self.env['TERM'] == self.TTYPE_UNDETECTED)
                if (not pair[0] in self.env or overwrite):
                    self.log.debug('env[%r] = %r', pair[0], pair[1])
                    self.env[pair[0]] = pair[1]
                elif pair[1] == self.env[pair[0]]:
                    self.log.debug('env[%r] repeated', pair[0])
                else:
                    self.log.warn('%s=%s; conflicting value %s ignored.',
                                  pair[0], self.env[pair[0]], pair[1])
            else:
                self.log.error('client NEW_ENVIRON; invalid %r', pair)
        self.ENV_REPLIED = True

    def _sb_naws(self, charbuf):
        """
        Processes incoming subnegotiation NAWS
        """
        if 5 != len(charbuf):
            self.log.error('{self.addrport}: bad length in NAWS buf ({buflen})'
                           .format(self=self, buflen=len(charbuf)))
            return

        columns = (256 * ord(charbuf[1])) + ord(charbuf[2])
        rows = (256 * ord(charbuf[3])) + ord(charbuf[4])
        old_rows = self.env.get('LINES', None)
        old_columns = self.env.get('COLUMNS', None)
        if (old_rows == str(rows) and old_columns == str(columns)):
            self.log.debug('{self.addrport}: NAWS repeated'.format(self=self))
            return
        if rows <= 0:
            self.log.debug('LINES %s ignored', rows)
            rows = old_rows
        if columns <= 0:
            self.log.debug('COLUMNS %s ignored', columns)
            columns = old_columns
        self.env['LINES'] = str(rows)
        self.env['COLUMNS'] = str(columns)
        if self.on_naws is not None:
            self.on_naws(self)

    #---[ State Juggling for Telnet Options ]----------------------------------
    def check_local_option(self, option):
        """
        Test the status of local negotiated Telnet options.
        """
        if option not in self.telnet_opt_dict:
            self.telnet_opt_dict[option] = TelnetOption()
        return self.telnet_opt_dict[option].local_option

    def _note_local_option(self, option, state):
        """
        Record the status of local negotiated Telnet options.
        """
        if option not in self.telnet_opt_dict:
            self.telnet_opt_dict[option] = TelnetOption()
        self.telnet_opt_dict[option].local_option = state

    def check_remote_option(self, option):
        """
        Test the status of remote negotiated Telnet options.
        """
        if option not in self.telnet_opt_dict:
            self.telnet_opt_dict[option] = TelnetOption()
        return self.telnet_opt_dict[option].remote_option

    def _note_remote_option(self, option, state):
        """
        Record the status of local negotiated Telnet options.
        """
        if option not in self.telnet_opt_dict:
            self.telnet_opt_dict[option] = TelnetOption()
        self.telnet_opt_dict[option].remote_option = state

    def _check_reply_pending(self, option):
        """
        Test the status of requested Telnet options.
        """
        if option not in self.telnet_opt_dict:
            self.telnet_opt_dict[option] = TelnetOption()
        return self.telnet_opt_dict[option].reply_pending

    def _note_reply_pending(self, option, state):
        """
        Record the status of requested Telnet options.
        """
        if option not in self.telnet_opt_dict:
            self.telnet_opt_dict[option] = TelnetOption()
        self.telnet_opt_dict[option].reply_pending = state

    #---[ Telnet Command Shortcuts ]-------------------------------------------
    def _iac_do(self, option):
        """
        Send a Telnet IAC "DO" sequence.
        """
        self.log.debug('send IAC DO %s', name_option(option))
        self.send_str(bytes(''.join((IAC, DO, option))))

    def _iac_dont(self, option):
        """
        Send a Telnet IAC "DONT" sequence.
        """
        self.log.debug('send IAC DONT %s', name_option(option))
        self.send_str(bytes(''.join((IAC, DONT, option))))

    def _iac_will(self, option):
        """
        Send a Telnet IAC "WILL" sequence.
        """
        self.log.debug('send IAC WILL %s', name_option(option))
        self.send_str(bytes(''.join((IAC, WILL, option))))

    def _iac_wont(self, option):
        """
        Send a Telnet IAC "WONT" sequence.
        """
        self.log.debug('send IAC WONT %s', name_option(option))
        self.send_str(bytes(''.join((IAC, WONT, option))))


class ConnectTelnet(BaseConnect):

    """
    Accept new Telnet Connection and negotiate options.
    """
    #: maximum time elapsed allowed to begin on-connect negotiation
    TIME_NEGOTIATE = 2.50
    #: wait upto 3500ms for all stages of negotiation to complete
    TIME_WAIT_STAGE = 3.50
    #: polling duration during negotiation
    TIME_POLL = 0.10

    def banner(self):
        """
        This method is called after the connection is initiated.

        This routine happens to communicate with a wide variety of network
        scanners when listening on the default port on a public IP address.
        """
        # According to Roger Espel Llima (espel@drakkar.ens.fr), you can
        #   have your server send a sequence of control characters:
        # (0xff 0xfb 0x01) (0xff 0xfb 0x03) (0xff 0xfd 0x0f3).
        #   Which translates to:
        # (IAC WILL ECHO) (IAC WILL SUPPRESS-GO-AHEAD)
        # (IAC DO SUPPRESS-GO-AHEAD).
        self.client.request_will_echo()
        self.client.request_will_sga()
        self.client.request_do_sga()
        # add DO & WILL BINARY, for utf8 input/output.
        self.client.request_do_binary()
        self.client.request_will_binary()
        # and terminal type, naws, and env,
        self.client.request_do_ttype()
        self.client.request_do_naws()
        self.client.request_do_env()
        self.client.send()  # push

    def run(self):
        """
        Negotiate and inquire about terminal type, telnet options, window size,
        and tcp socket options before spawning a new session.
        """
        try:
            self._set_socket_opts()
            mrk_bytes = self.client.bytes_received

            self.banner()

            # wait at least {TIME_NEGOTIATE} for the client to speak.
            # If it doesn't, we try only TTYPE.
            # If it fails to report that, we forget the rest.
            self.log.debug('{client.addrport}: pausing for negotiation'
                           .format(client=self.client))
            st_time = time.time()
            while ((0 == self.client.bytes_received
                    or mrk_bytes == self.client.bytes_received)
                   and time.time() - st_time < self.TIME_NEGOTIATE):
                if not self.client.is_active():
                    return
                if self.client.send_ready():
                    self.client.send()
                time.sleep(self.TIME_POLL)

            if self._check_ttype(start_time=st_time):
                # if client is able to negotiate terminal type,
                # try NAWS or ENV negotiation.
                if self.client.is_active():
                    self._check_env(start_time=st_time)
                if self.client.is_active():
                    self._check_naws(start_time=st_time)

            self.set_encoding()

            if self.client.is_active():
                return spawn_client_session(client=self.client)
        except (Disconnected, socket.error) as err:
            self.log.debug('{client.addrport}: connection closed: {err}'
                           .format(client=self.client, err=err))
        except EOFError:
            self.log.debug('{client.addrport}: EOF from client'
                           .format(client=self.client))
        except Exception as err:
            self.log.debug('{client.addrport}: connection closed: {err}'
                           .format(client=self.client, err=err))
        finally:
            self.stopped = True
        self.client.deactivate()

    def set_encoding(self):
        # set encoding to utf8 for clients negotiating BINARY mode and
        # not beginning with TERM 'ansi'.
        #
        # This assumes a very simple dualistic model: modern unix terminal
        # (probably using bsd telnet client), or SyncTerm.
        #
        # Clients *could* negotiate CHARSET for encoding, or simply forward a
        # LANG variable -- SyncTerm does neither. So we just assume any
        # terminal that says its "ansi" is just any number of old-world DOS
        # terminal emulating clients that are incapable of comprehending
        # "encoding" (Especially multi-byte!), they only decide upon a "font"
        # or "code page" to map char 0-255 to.
        #
        self.client.env['encoding'] = 'cp437'
        local = self.client.check_local_option
        remote = self.client.check_remote_option
        term = self.client.env.get('TERM', 'ansi')

        if (local(BINARY) and remote(BINARY) and not term.startswith('ansi')):
            self.client.env['encoding'] = 'utf8'

    def _timeleft(self, st_time):
        """
        Returns True when difference of current time and st_time is below
        TIME_WAIT_STAGE.
        """
        return bool(time.time() - st_time < self.TIME_WAIT_STAGE)

    def _check_env(self, start_time):
        """
        Check for NEW_ENVIRON negotiation.
        """
        def detected():
            return (self.client.check_remote_option(NEW_ENVIRON) is not UNKNOWN
                    or self.client.ENV_REPLIED)

        while not detected() and self._timeleft(start_time):
            if not self.client.is_active():
                return
            if self.client.send_ready():
                self.client.send()
            time.sleep(self.TIME_POLL)

        if detected():
            self.log.debug('{client.addrport}: ENV={client.env!r}.'
                           .format(client=self.client))
        else:
            self.log.debug('{client.addrport}: request-do-new_environ failed.'
                           .format(client=self.client))

    def _check_naws(self, start_time):
        """
        Check for NAWS negotiation.
        """
        def detected():
            return (self.client.env.get('LINES', None) is not None
                    and self.client.env.get('COLUMNS', None) is not None)

        while not detected() and self._timeleft(start_time):
            if not self.client.is_active():
                return
            if self.client.send_ready():
                self.client.send()
            time.sleep(self.TIME_POLL)

        if detected():
            self.log.debug('{client.addrport}: COLUMNS={client.env[COLUMNS]}, '
                           'LINES={client.env[LINES]}.'
                           .format(client=self.client))
        else:
            self.log.debug('{client.addrport}: request-do-naws failed.'
                           .format(client=self.client))

    def _check_ttype(self, start_time):
        """
        Check for TTYPE negotiation.
        """
        def detected():
            return self.client.env['TERM'] != self.client.TTYPE_UNDETECTED

        while not detected() and self._timeleft(start_time):
            if not self.client.is_active():
                return False
            if self.client.send_ready():
                self.client.send()
            time.sleep(self.TIME_POLL)

        if detected():
            self.log.debug('{client.addrport}: TERM={client.env[TERM]}.'
                           .format(client=self.client))
            return True
        self.log.debug('{client.addrport}: request-terminal-type failed.'
                       .format(client=self.client))
        return False


class TelnetServer(BaseServer):

    """
    Poll sockets for new connections and sending/receiving data from clients.
    """

    client_factory = TelnetClient
    connect_factory = ConnectTelnet
    client_factory_kwargs = dict(on_naws=on_naws)

    # Dictionary of active clients, (file descriptor, TelnetClient,)
    clients = {}

    def __init__(self, config):
        """
        Create a new Telnet Server.

        :param ConfigParser.ConfigParser config: configuration section
                                         ``[telnet]``, with options ``'addr'``,
                                         ``'port'``
        """
        self.log = logging.getLogger(__name__)
        self.address = config.get('telnet', 'addr')
        self.port = config.getint('telnet', 'port')

        # bind
        self.server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.server_socket.setsockopt(
            socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        try:
            self.server_socket.bind((self.address, self.port))
            self.server_socket.listen(self.LISTEN_BACKLOG)
        except socket.error as err:
            self.log.error('Unable to bind {0}:{1}: {2}'
                           .format(self.address, self.port, err))
            exit(1)
        self.log.info('telnet listening on {self.address}:{self.port}/tcp'
                      .format(self=self))



""" Main menu script for x/84. """
# std imports
from __future__ import division
import collections
import os

# local
from x84.bbs import (
    syncterm_setfont,
    getterminal,
    getsession,
    LineEditor,
    get_ini,
    gosub,
    echo,
    ini,
)
from common import (
    render_menu_entries,
    display_banner,
    display_prompt,
)

#: MenuItem is a definition class for display, input, and target script.
MenuItem = collections.namedtuple(
    'MenuItem', ['inp_key', 'text', 'script', 'args', 'kwargs'])

#: When set False, menu items are not colorized and render much
#: faster on slower systems (such as raspberry pi).
colored_menu_items = get_ini(
    section='main', key='colored_menu_items', getter='getboolean'
) or True

#: color used for menu key entries
color_highlight = get_ini(
    section='main', key='color_highlight'
) or 'bold_magenta'

#: color used for prompt
color_backlight = get_ini(
    section='main', key='color_backlight',
) or 'magenta_reverse'

#: color used for brackets ``[`` and ``]``
color_lowlight = get_ini(
    section='main', key='color_lowlight'
) or 'bold_black'

#: filepath to artfile displayed for this script
art_file = get_ini(
    section='main', key='art_file'
) or 'art/main1.asc'

#: encoding used to display artfile
art_encoding = get_ini(
    section='main', key='art_encoding'
) or 'cp437'  # ascii, actually

#: fontset for SyncTerm emulator
syncterm_font = get_ini(
    section='main', key='syncterm_font'
) or 'topaz'


def get_sesame_menu_items(session):
    # there doesn't exist any documentation on how this works,
    # only the given examples in the generated default.ini file
    menu_items = []
    if ini.CFG.has_section('sesame'):
        for name in filter(lambda _name: '_' not in _name,
                           ini.CFG.options('sesame')):

            sesame_kwargs = {'name': name}

            door_cmd = ini.CFG.get('sesame', name).split(None, 1)[0]
            if door_cmd.lower() == 'no' or not os.path.exists(door_cmd):
                # skip entry if path does not resolve, or set to 'no'
                continue

            inp_key = get_ini(section='sesame', key='{0}_key'.format(name))
            if not inp_key:
                raise ValueError('sesame configuration for "{0}" requires '
                                 'complimenting value "{0}_key" for menu '
                                 'input key.'.format(name))

            if get_ini(section='sesame', key='{0}_sysop_only'.format(name),
                       getter='getboolean') and not session.user.is_sysop:
                continue

            text = get_ini(
                section='sesame', key='{0}_text'.format(name)
            ) or name

            menu_items.append(
                MenuItem(inp_key=inp_key, text=text, script='sesame',
                         args=(), kwargs=sesame_kwargs))

    return menu_items


def get_menu_items(session):
    """ Returns list of MenuItem entries. """
    #: A declaration of menu items and their acting gosub script
    menu_items = [
        # most 'expressive' scripts,
        MenuItem(inp_key=u'irc',
                 text=u'irc chat',
                 script='ircchat',
                 args=(), kwargs={}),
        MenuItem(inp_key=u'who',
                 text=u"who's online",
                 script='online',
                 args=(), kwargs={}),
        MenuItem(inp_key=u'fb',
                 text=u'file browser',
                 script='fbrowse',
                 args=(), kwargs={}),
        MenuItem(inp_key=u'pe',
                 text=u'profile editor',
                 script='profile',
                 args=(), kwargs={}),
        MenuItem(inp_key=u'weather',
                 text=u'weather forecast',
                 script='weather',
                 args=(), kwargs={}),
        MenuItem(inp_key=u'hn',
                 text=u'hacker news',
                 script='hackernews',
                 args=(), kwargs={}),
        MenuItem(inp_key=u'ol',
                 text=u'one-liners',
                 script='ol',
                 args=(), kwargs={}),
        MenuItem(inp_key=u'tetris',
                 text=u'tetris game',
                 script='tetris',
                 args=(), kwargs={}),
        MenuItem(inp_key=u'vote',
                 text=u'voting booth',
                 script='vote',
                 args=(), kwargs={}),
        MenuItem(inp_key=u'lc',
                 text=u'last callers',
                 script='lc',
                 args=(), kwargs={}),
        MenuItem(inp_key=u'user',
                 text=u'user list',
                 script='userlist',
                 args=(), kwargs={}),
        MenuItem(inp_key=u'news',
                 text=u'news reader',
                 script='news',
                 args=(), kwargs={}),
        MenuItem(inp_key=u'si',
                 text=u'system info',
                 script='si',
                 args=(), kwargs={}),
        MenuItem(inp_key=u'key',
                 text=u'keyboard test',
                 script='test_keyboard_keys',
                 args=(), kwargs={}),
        MenuItem(inp_key=u'ac',
                 text=u'adjust charset',
                 script='charset',
                 args=(), kwargs={}),

        MenuItem(inp_key=u'msg',
                 text=u'message area',
                 script='msgarea',
                 args=(), kwargs={}),

        MenuItem(inp_key=u'g',
                 text=u'logoff system',
                 script='logoff',
                 args=(), kwargs={}),

    ]

    # add sysop menu for sysop users, only.
    if session.user.is_sysop:
        menu_items.append(
            MenuItem(inp_key='sysop',
                     text=u'sysop area',
                     script='sysop',
                     args=(), kwargs={}))

    # add sesame doors, if any.
    menu_items.extend(get_sesame_menu_items(session))
    return menu_items


def get_line_editor(term, menu):
    """ Return a line editor suitable for menu entry prompts. """
    # if inp_key's were CJK characters, you should use term.length to measure
    # printable length of double-wide characters ... this is too costly to
    # enable by default.  Just a note for you east-asian folks.
    max_inp_length = max([len(item.inp_key) for item in menu])
    return LineEditor(width=max_inp_length,
                      colors={'highlight': getattr(term, color_backlight)})


def main():
    """ Main menu entry point. """
    session, term = getsession(), getterminal()

    text, width, height, dirty = u'', -1, -1, 2
    menu_items = get_menu_items(session)
    editor = get_line_editor(term, menu_items)
    colors = {}
    if colored_menu_items:
        colors['backlight'] = getattr(term, color_backlight)
        colors['highlight'] = getattr(term, color_highlight)
        colors['lowlight'] = getattr(term, color_lowlight)

    while True:
        if dirty == 2:
            # set syncterm font, if any
            if syncterm_font and term.kind.startswith('ansi'):
                echo(syncterm_setfont(syncterm_font))
        if dirty:
            session.activity = 'main menu'
            top_margin = display_banner(art_file, encoding=art_encoding) + 1
            echo(u'\r\n')
            if width != term.width or height != term.height:
                width, height = term.width, term.height
                text = render_menu_entries(
                    term, top_margin, menu_items, colors)
            echo(u''.join((text,
                           display_prompt(term, colors),
                           editor.refresh())))
            dirty = 0

        event, data = session.read_events(('input', 'refresh'))

        if event == 'refresh':
            dirty = True
            continue

        elif event == 'input':
            session.buffer_input(data, pushback=True)

            # we must loop over inkey(0), we received a 'data'
            # event, though there may be many keystrokes awaiting for our
            # decoding -- or none at all (multibyte sequence not yet complete).
            inp = term.inkey(0)
            while inp:
                if inp.code == term.KEY_ENTER:
                    # find matching menu item,
                    for item in menu_items:
                        if item.inp_key == editor.content.strip():
                            echo(term.normal + u'\r\n')
                            gosub(item.script, *item.args, **item.kwargs)
                            editor.content = u''
                            dirty = 2
                            break
                    else:
                        if editor.content:
                            # command not found, clear prompt.
                            echo(u''.join((
                                (u'\b' * len(editor.content)),
                                (u' ' * len(editor.content)),
                                (u'\b' * len(editor.content)),)))
                            editor.content = u''
                            echo(editor.refresh())
                elif inp.is_sequence:
                    echo(editor.process_keystroke(inp.code))
                else:
                    echo(editor.process_keystroke(inp))
                inp = term.inkey(0)



"""
IBM PC Code Page 437 codec for x/84.

Generated from:
ftp://ftp.unicode.org/Public/MAPPINGS/VENDORS/MICSFT/PC/CP437.TXT
"""

import codecs

# Codec APIs


class Codec(codecs.Codec):

    def encode(self, char, errors='strict'):
        return codecs.charmap_encode(char, errors, ENCODING_MAP)

    def decode(self, char, errors='strict'):
        return codecs.charmap_decode(char, errors, DECODING_TABLE)


class IncrementalEncoder(codecs.IncrementalEncoder):

    def encode(self, char, final=False):
        return codecs.charmap_encode(char, self.errors, ENCODING_MAP)[0]


class IncrementalDecoder(codecs.IncrementalDecoder):

    def decode(self, char, final=False):
        return codecs.charmap_decode(char, self.errors, DECODING_TABLE)[0]


class StreamWriter(Codec, codecs.StreamWriter):
    pass


class StreamReader(Codec, codecs.StreamReader):
    pass

# encodings module API


def getaliases():
    return (
        'ibmpc',
        'ibm_pc',
        'msdos',
        'pc',
    )


def getregentry():
    return codecs.CodecInfo(
        name='cp437',
        encode=Codec().encode,
        decode=Codec().decode,
        incrementalencoder=IncrementalEncoder,
        incrementaldecoder=IncrementalDecoder,
        streamreader=StreamReader,
        streamwriter=StreamWriter,
    )

# Decoding Map

DECODING_MAP = codecs.make_identity_dict(range(256))
DECODING_MAP.update({
    0x0080: 0x00c7,  # LATIN CAPITAL LETTER C WITH CEDILLA
    0x0081: 0x00fc,  # LATIN SMALL LETTER U WITH DIAERESIS
    0x0082: 0x00e9,  # LATIN SMALL LETTER E WITH ACUTE
    0x0083: 0x00e2,  # LATIN SMALL LETTER A WITH CIRCUMFLEX
    0x0084: 0x00e4,  # LATIN SMALL LETTER A WITH DIAERESIS
    0x0085: 0x00e0,  # LATIN SMALL LETTER A WITH GRAVE
    0x0086: 0x00e5,  # LATIN SMALL LETTER A WITH RING ABOVE
    0x0087: 0x00e7,  # LATIN SMALL LETTER C WITH CEDILLA
    0x0088: 0x00ea,  # LATIN SMALL LETTER E WITH CIRCUMFLEX
    0x0089: 0x00eb,  # LATIN SMALL LETTER E WITH DIAERESIS
    0x008a: 0x00e8,  # LATIN SMALL LETTER E WITH GRAVE
    0x008b: 0x00ef,  # LATIN SMALL LETTER I WITH DIAERESIS
    0x008c: 0x00ee,  # LATIN SMALL LETTER I WITH CIRCUMFLEX
    0x008d: 0x00ec,  # LATIN SMALL LETTER I WITH GRAVE
    0x008e: 0x00c4,  # LATIN CAPITAL LETTER A WITH DIAERESIS
    0x008f: 0x00c5,  # LATIN CAPITAL LETTER A WITH RING ABOVE
    0x0090: 0x00c9,  # LATIN CAPITAL LETTER E WITH ACUTE
    0x0091: 0x00e6,  # LATIN SMALL LIGATURE AE
    0x0092: 0x00c6,  # LATIN CAPITAL LIGATURE AE
    0x0093: 0x00f4,  # LATIN SMALL LETTER O WITH CIRCUMFLEX
    0x0094: 0x00f6,  # LATIN SMALL LETTER O WITH DIAERESIS
    0x0095: 0x00f2,  # LATIN SMALL LETTER O WITH GRAVE
    0x0096: 0x00fb,  # LATIN SMALL LETTER U WITH CIRCUMFLEX
    0x0097: 0x00f9,  # LATIN SMALL LETTER U WITH GRAVE
    0x0098: 0x00ff,  # LATIN SMALL LETTER Y WITH DIAERESIS
    0x0099: 0x00d6,  # LATIN CAPITAL LETTER O WITH DIAERESIS
    0x009a: 0x00dc,  # LATIN CAPITAL LETTER U WITH DIAERESIS
    0x009b: 0x00a2,  # CENT SIGN
    0x009c: 0x00a3,  # POUND SIGN
    0x009d: 0x00a5,  # YEN SIGN
    0x009e: 0x20a7,  # PESETA SIGN
    0x009f: 0x0192,  # LATIN SMALL LETTER F WITH HOOK
    0x00a0: 0x00e1,  # LATIN SMALL LETTER A WITH ACUTE
    0x00a1: 0x00ed,  # LATIN SMALL LETTER I WITH ACUTE
    0x00a2: 0x00f3,  # LATIN SMALL LETTER O WITH ACUTE
    0x00a3: 0x00fa,  # LATIN SMALL LETTER U WITH ACUTE
    0x00a4: 0x00f1,  # LATIN SMALL LETTER N WITH TILDE
    0x00a5: 0x00d1,  # LATIN CAPITAL LETTER N WITH TILDE
    0x00a6: 0x00aa,  # FEMININE ORDINAL INDICATOR
    0x00a7: 0x00ba,  # MASCULINE ORDINAL INDICATOR
    0x00a8: 0x00bf,  # INVERTED QUESTION MARK
    0x00a9: 0x2310,  # REVERSED NOT SIGN
    0x00aa: 0x00ac,  # NOT SIGN
    0x00ab: 0x00bd,  # VULGAR FRACTION ONE HALF
    0x00ac: 0x00bc,  # VULGAR FRACTION ONE QUARTER
    0x00ad: 0x00a1,  # INVERTED EXCLAMATION MARK
    0x00ae: 0x00ab,  # LEFT-POINTING DOUBLE ANGLE QUOTATION MARK
    0x00af: 0x00bb,  # RIGHT-POINTING DOUBLE ANGLE QUOTATION MARK
    0x00b0: 0x2591,  # LIGHT SHADE
    0x00b1: 0x2592,  # MEDIUM SHADE
    0x00b2: 0x2593,  # DARK SHADE
    0x00b3: 0x2502,  # BOX DRAWINGS LIGHT VERTICAL
    0x00b4: 0x2524,  # BOX DRAWINGS LIGHT VERTICAL AND LEFT
    0x00b5: 0x2561,  # BOX DRAWINGS VERTICAL SINGLE AND LEFT DOUBLE
    0x00b6: 0x2562,  # BOX DRAWINGS VERTICAL DOUBLE AND LEFT SINGLE
    0x00b7: 0x2556,  # BOX DRAWINGS DOWN DOUBLE AND LEFT SINGLE
    0x00b8: 0x2555,  # BOX DRAWINGS DOWN SINGLE AND LEFT DOUBLE
    0x00b9: 0x2563,  # BOX DRAWINGS DOUBLE VERTICAL AND LEFT
    0x00ba: 0x2551,  # BOX DRAWINGS DOUBLE VERTICAL
    0x00bb: 0x2557,  # BOX DRAWINGS DOUBLE DOWN AND LEFT
    0x00bc: 0x255d,  # BOX DRAWINGS DOUBLE UP AND LEFT
    0x00bd: 0x255c,  # BOX DRAWINGS UP DOUBLE AND LEFT SINGLE
    0x00be: 0x255b,  # BOX DRAWINGS UP SINGLE AND LEFT DOUBLE
    0x00bf: 0x2510,  # BOX DRAWINGS LIGHT DOWN AND LEFT
    0x00c0: 0x2514,  # BOX DRAWINGS LIGHT UP AND RIGHT
    0x00c1: 0x2534,  # BOX DRAWINGS LIGHT UP AND HORIZONTAL
    0x00c2: 0x252c,  # BOX DRAWINGS LIGHT DOWN AND HORIZONTAL
    0x00c3: 0x251c,  # BOX DRAWINGS LIGHT VERTICAL AND RIGHT
    0x00c4: 0x2500,  # BOX DRAWINGS LIGHT HORIZONTAL
    0x00c5: 0x253c,  # BOX DRAWINGS LIGHT VERTICAL AND HORIZONTAL
    0x00c6: 0x255e,  # BOX DRAWINGS VERTICAL SINGLE AND RIGHT DOUBLE
    0x00c7: 0x255f,  # BOX DRAWINGS VERTICAL DOUBLE AND RIGHT SINGLE
    0x00c8: 0x255a,  # BOX DRAWINGS DOUBLE UP AND RIGHT
    0x00c9: 0x2554,  # BOX DRAWINGS DOUBLE DOWN AND RIGHT
    0x00ca: 0x2569,  # BOX DRAWINGS DOUBLE UP AND HORIZONTAL
    0x00cb: 0x2566,  # BOX DRAWINGS DOUBLE DOWN AND HORIZONTAL
    0x00cc: 0x2560,  # BOX DRAWINGS DOUBLE VERTICAL AND RIGHT
    0x00cd: 0x2550,  # BOX DRAWINGS DOUBLE HORIZONTAL
    0x00ce: 0x256c,  # BOX DRAWINGS DOUBLE VERTICAL AND HORIZONTAL
    0x00cf: 0x2567,  # BOX DRAWINGS UP SINGLE AND HORIZONTAL DOUBLE
    0x00d0: 0x2568,  # BOX DRAWINGS UP DOUBLE AND HORIZONTAL SINGLE
    0x00d1: 0x2564,  # BOX DRAWINGS DOWN SINGLE AND HORIZONTAL DOUBLE
    0x00d2: 0x2565,  # BOX DRAWINGS DOWN DOUBLE AND HORIZONTAL SINGLE
    0x00d3: 0x2559,  # BOX DRAWINGS UP DOUBLE AND RIGHT SINGLE
    0x00d4: 0x2558,  # BOX DRAWINGS UP SINGLE AND RIGHT DOUBLE
    0x00d5: 0x2552,  # BOX DRAWINGS DOWN SINGLE AND RIGHT DOUBLE
    0x00d6: 0x2553,  # BOX DRAWINGS DOWN DOUBLE AND RIGHT SINGLE
    0x00d7: 0x256b,  # BOX DRAWINGS VERTICAL DOUBLE AND HORIZONTAL SINGLE
    0x00d8: 0x256a,  # BOX DRAWINGS VERTICAL SINGLE AND HORIZONTAL DOUBLE
    0x00d9: 0x2518,  # BOX DRAWINGS LIGHT UP AND LEFT
    0x00da: 0x250c,  # BOX DRAWINGS LIGHT DOWN AND RIGHT
    0x00db: 0x2588,  # FULL BLOCK
    0x00dc: 0x2584,  # LOWER HALF BLOCK
    0x00dd: 0x258c,  # LEFT HALF BLOCK
    0x00de: 0x2590,  # RIGHT HALF BLOCK
    0x00df: 0x2580,  # UPPER HALF BLOCK
    0x00e0: 0x03b1,  # GREEK SMALL LETTER ALPHA
    0x00e1: 0x00df,  # LATIN SMALL LETTER SHARP S
    0x00e2: 0x0393,  # GREEK CAPITAL LETTER GAMMA
    0x00e3: 0x03c0,  # GREEK SMALL LETTER PI
    0x00e4: 0x03a3,  # GREEK CAPITAL LETTER SIGMA
    0x00e5: 0x03c3,  # GREEK SMALL LETTER SIGMA
    0x00e6: 0x00b5,  # MICRO SIGN
    0x00e7: 0x03c4,  # GREEK SMALL LETTER TAU
    0x00e8: 0x03a6,  # GREEK CAPITAL LETTER PHI
    0x00e9: 0x0398,  # GREEK CAPITAL LETTER THETA
    0x00ea: 0x03a9,  # GREEK CAPITAL LETTER OMEGA
    0x00eb: 0x03b4,  # GREEK SMALL LETTER DELTA
    0x00ec: 0x221e,  # INFINITY
    0x00ed: 0x03c6,  # GREEK SMALL LETTER PHI
    0x00ee: 0x03b5,  # GREEK SMALL LETTER EPSILON
    0x00ef: 0x2229,  # INTERSECTION
    0x00f0: 0x2261,  # IDENTICAL TO
    0x00f1: 0x00b1,  # PLUS-MINUS SIGN
    0x00f2: 0x2265,  # GREATER-THAN OR EQUAL TO
    0x00f3: 0x2264,  # LESS-THAN OR EQUAL TO
    0x00f4: 0x2320,  # TOP HALF INTEGRAL
    0x00f5: 0x2321,  # BOTTOM HALF INTEGRAL
    0x00f6: 0x00f7,  # DIVISION SIGN
    0x00f7: 0x2248,  # ALMOST EQUAL TO
    0x00f8: 0x00b0,  # DEGREE SIGN
    0x00f9: 0x2219,  # BULLET OPERATOR
    0x00fa: 0x00b7,  # MIDDLE DOT
    0x00fb: 0x221a,  # SQUARE ROOT
    0x00fc: 0x207f,  # SUPERSCRIPT LATIN SMALL LETTER N
    0x00fd: 0x00b2,  # SUPERSCRIPT TWO
    0x00fe: 0x25a0,  # BLACK SQUARE
    0x00ff: 0x00a0,  # NO-BREAK SPACE
})

# Decoding Table

DECODING_TABLE = (
    u'\x00'  # 0x0000 -> NULL
    u'\x01'  # 0x0001 -> START OF HEADING
    u'\x02'  # 0x0002 -> START OF TEXT
    u'\x03'  # 0x0003 -> END OF TEXT
    u'\x04'  # 0x0004 -> END OF TRANSMISSION
    u'\x05'  # 0x0005 -> ENQUIRY
    u'\x06'  # 0x0006 -> ACKNOWLEDGE
    u'\x07'  # 0x0007 -> BELL
    u'\x08'  # 0x0008 -> BACKSPACE
    u'\t'  # 0x0009 -> HORIZONTAL TABULATION
    u'\n'  # 0x000a -> LINE FEED
    u'\x0b'  # 0x000b -> VERTICAL TABULATION
    u'\x0c'  # 0x000c -> FORM FEED
    u'\r'  # 0x000d -> CARRIAGE RETURN
    u'\x0e'  # 0x000e -> SHIFT OUT
    u'\x0f'  # 0x000f -> SHIFT IN
    u'\x10'  # 0x0010 -> DATA LINK ESCAPE
    u'\x11'  # 0x0011 -> DEVICE CONTROL ONE
    u'\x12'  # 0x0012 -> DEVICE CONTROL TWO
    u'\x13'  # 0x0013 -> DEVICE CONTROL THREE
    u'\x14'  # 0x0014 -> DEVICE CONTROL FOUR
    u'\x15'  # 0x0015 -> NEGATIVE ACKNOWLEDGE
    u'\x16'  # 0x0016 -> SYNCHRONOUS IDLE
    u'\x17'  # 0x0017 -> END OF TRANSMISSION BLOCK
    u'\x18'  # 0x0018 -> CANCEL
    u'\x19'  # 0x0019 -> END OF MEDIUM
    u'\x1a'  # 0x001a -> SUBSTITUTE
    u'\x1b'  # 0x001b -> ESCAPE
    u'\x1c'  # 0x001c -> FILE SEPARATOR
    u'\x1d'  # 0x001d -> GROUP SEPARATOR
    u'\x1e'  # 0x001e -> RECORD SEPARATOR
    u'\x1f'  # 0x001f -> UNIT SEPARATOR
    u' '  # 0x0020 -> SPACE
    u'!'  # 0x0021 -> EXCLAMATION MARK
    u'"'  # 0x0022 -> QUOTATION MARK
    u'#'  # 0x0023 -> NUMBER SIGN
    u'$'  # 0x0024 -> DOLLAR SIGN
    u'%'  # 0x0025 -> PERCENT SIGN
    u'&'  # 0x0026 -> AMPERSAND
    u"'"  # 0x0027 -> APOSTROPHE
    u'('  # 0x0028 -> LEFT PARENTHESIS
    u')'  # 0x0029 -> RIGHT PARENTHESIS
    u'*'  # 0x002a -> ASTERISK
    u'+'  # 0x002b -> PLUS SIGN
    u','  # 0x002c -> COMMA
    u'-'  # 0x002d -> HYPHEN-MINUS
    u'.'  # 0x002e -> FULL STOP
    u'/'  # 0x002f -> SOLIDUS
    u'0'  # 0x0030 -> DIGIT ZERO
    u'1'  # 0x0031 -> DIGIT ONE
    u'2'  # 0x0032 -> DIGIT TWO
    u'3'  # 0x0033 -> DIGIT THREE
    u'4'  # 0x0034 -> DIGIT FOUR
    u'5'  # 0x0035 -> DIGIT FIVE
    u'6'  # 0x0036 -> DIGIT SIX
    u'7'  # 0x0037 -> DIGIT SEVEN
    u'8'  # 0x0038 -> DIGIT EIGHT
    u'9'  # 0x0039 -> DIGIT NINE
    u':'  # 0x003a -> COLON
    u';'  # 0x003b -> SEMICOLON
    u'<'  # 0x003c -> LESS-THAN SIGN
    u'='  # 0x003d -> EQUALS SIGN
    u'>'  # 0x003e -> GREATER-THAN SIGN
    u'?'  # 0x003f -> QUESTION MARK
    u'@'  # 0x0040 -> COMMERCIAL AT
    u'A'  # 0x0041 -> LATIN CAPITAL LETTER A
    u'B'  # 0x0042 -> LATIN CAPITAL LETTER B
    u'C'  # 0x0043 -> LATIN CAPITAL LETTER C
    u'D'  # 0x0044 -> LATIN CAPITAL LETTER D
    u'E'  # 0x0045 -> LATIN CAPITAL LETTER E
    u'F'  # 0x0046 -> LATIN CAPITAL LETTER F
    u'G'  # 0x0047 -> LATIN CAPITAL LETTER G
    u'H'  # 0x0048 -> LATIN CAPITAL LETTER H
    u'I'  # 0x0049 -> LATIN CAPITAL LETTER I
    u'J'  # 0x004a -> LATIN CAPITAL LETTER J
    u'K'  # 0x004b -> LATIN CAPITAL LETTER K
    u'L'  # 0x004c -> LATIN CAPITAL LETTER L
    u'M'  # 0x004d -> LATIN CAPITAL LETTER M
    u'N'  # 0x004e -> LATIN CAPITAL LETTER N
    u'O'  # 0x004f -> LATIN CAPITAL LETTER O
    u'P'  # 0x0050 -> LATIN CAPITAL LETTER P
    u'Q'  # 0x0051 -> LATIN CAPITAL LETTER Q
    u'R'  # 0x0052 -> LATIN CAPITAL LETTER R
    u'S'  # 0x0053 -> LATIN CAPITAL LETTER S
    u'T'  # 0x0054 -> LATIN CAPITAL LETTER T
    u'U'  # 0x0055 -> LATIN CAPITAL LETTER U
    u'V'  # 0x0056 -> LATIN CAPITAL LETTER V
    u'W'  # 0x0057 -> LATIN CAPITAL LETTER W
    u'X'  # 0x0058 -> LATIN CAPITAL LETTER X
    u'Y'  # 0x0059 -> LATIN CAPITAL LETTER Y
    u'Z'  # 0x005a -> LATIN CAPITAL LETTER Z
    u'['  # 0x005b -> LEFT SQUARE BRACKET
    u'\\'  # 0x005c -> REVERSE SOLIDUS
    u']'  # 0x005d -> RIGHT SQUARE BRACKET
    u'^'  # 0x005e -> CIRCUMFLEX ACCENT
    u'_'  # 0x005f -> LOW LINE
    u'`'  # 0x0060 -> GRAVE ACCENT
    u'a'  # 0x0061 -> LATIN SMALL LETTER A
    u'b'  # 0x0062 -> LATIN SMALL LETTER B
    u'c'  # 0x0063 -> LATIN SMALL LETTER C
    u'd'  # 0x0064 -> LATIN SMALL LETTER D
    u'e'  # 0x0065 -> LATIN SMALL LETTER E
    u'f'  # 0x0066 -> LATIN SMALL LETTER F
    u'g'  # 0x0067 -> LATIN SMALL LETTER G
    u'h'  # 0x0068 -> LATIN SMALL LETTER H
    u'i'  # 0x0069 -> LATIN SMALL LETTER I
    u'j'  # 0x006a -> LATIN SMALL LETTER J
    u'k'  # 0x006b -> LATIN SMALL LETTER K
    u'l'  # 0x006c -> LATIN SMALL LETTER L
    u'm'  # 0x006d -> LATIN SMALL LETTER M
    u'n'  # 0x006e -> LATIN SMALL LETTER N
    u'o'  # 0x006f -> LATIN SMALL LETTER O
    u'p'  # 0x0070 -> LATIN SMALL LETTER P
    u'q'  # 0x0071 -> LATIN SMALL LETTER Q
    u'r'  # 0x0072 -> LATIN SMALL LETTER R
    u's'  # 0x0073 -> LATIN SMALL LETTER S
    u't'  # 0x0074 -> LATIN SMALL LETTER T
    u'u'  # 0x0075 -> LATIN SMALL LETTER U
    u'v'  # 0x0076 -> LATIN SMALL LETTER V
    u'w'  # 0x0077 -> LATIN SMALL LETTER W
    u'x'  # 0x0078 -> LATIN SMALL LETTER X
    u'y'  # 0x0079 -> LATIN SMALL LETTER Y
    u'z'  # 0x007a -> LATIN SMALL LETTER Z
    u'{'  # 0x007b -> LEFT CURLY BRACKET
    u'|'  # 0x007c -> VERTICAL LINE
    u'}'  # 0x007d -> RIGHT CURLY BRACKET
    u'~'  # 0x007e -> TILDE
    u'\x7f'  # 0x007f -> DELETE
    u'\xc7'  # 0x0080 -> LATIN CAPITAL LETTER C WITH CEDILLA
    u'\xfc'  # 0x0081 -> LATIN SMALL LETTER U WITH DIAERESIS
    u'\xe9'  # 0x0082 -> LATIN SMALL LETTER E WITH ACUTE
    u'\xe2'  # 0x0083 -> LATIN SMALL LETTER A WITH CIRCUMFLEX
    u'\xe4'  # 0x0084 -> LATIN SMALL LETTER A WITH DIAERESIS
    u'\xe0'  # 0x0085 -> LATIN SMALL LETTER A WITH GRAVE
    u'\xe5'  # 0x0086 -> LATIN SMALL LETTER A WITH RING ABOVE
    u'\xe7'  # 0x0087 -> LATIN SMALL LETTER C WITH CEDILLA
    u'\xea'  # 0x0088 -> LATIN SMALL LETTER E WITH CIRCUMFLEX
    u'\xeb'  # 0x0089 -> LATIN SMALL LETTER E WITH DIAERESIS
    u'\xe8'  # 0x008a -> LATIN SMALL LETTER E WITH GRAVE
    u'\xef'  # 0x008b -> LATIN SMALL LETTER I WITH DIAERESIS
    u'\xee'  # 0x008c -> LATIN SMALL LETTER I WITH CIRCUMFLEX
    u'\xec'  # 0x008d -> LATIN SMALL LETTER I WITH GRAVE
    u'\xc4'  # 0x008e -> LATIN CAPITAL LETTER A WITH DIAERESIS
    u'\xc5'  # 0x008f -> LATIN CAPITAL LETTER A WITH RING ABOVE
    u'\xc9'  # 0x0090 -> LATIN CAPITAL LETTER E WITH ACUTE
    u'\xe6'  # 0x0091 -> LATIN SMALL LIGATURE AE
    u'\xc6'  # 0x0092 -> LATIN CAPITAL LIGATURE AE
    u'\xf4'  # 0x0093 -> LATIN SMALL LETTER O WITH CIRCUMFLEX
    u'\xf6'  # 0x0094 -> LATIN SMALL LETTER O WITH DIAERESIS
    u'\xf2'  # 0x0095 -> LATIN SMALL LETTER O WITH GRAVE
    u'\xfb'  # 0x0096 -> LATIN SMALL LETTER U WITH CIRCUMFLEX
    u'\xf9'  # 0x0097 -> LATIN SMALL LETTER U WITH GRAVE
    u'\xff'  # 0x0098 -> LATIN SMALL LETTER Y WITH DIAERESIS
    u'\xd6'  # 0x0099 -> LATIN CAPITAL LETTER O WITH DIAERESIS
    u'\xdc'  # 0x009a -> LATIN CAPITAL LETTER U WITH DIAERESIS
    u'\xa2'  # 0x009b -> CENT SIGN
    u'\xa3'  # 0x009c -> POUND SIGN
    u'\xa5'  # 0x009d -> YEN SIGN
    u'\u20a7'  # 0x009e -> PESETA SIGN
    u'\u0192'  # 0x009f -> LATIN SMALL LETTER F WITH HOOK
    u'\xe1'  # 0x00a0 -> LATIN SMALL LETTER A WITH ACUTE
    u'\xed'  # 0x00a1 -> LATIN SMALL LETTER I WITH ACUTE
    u'\xf3'  # 0x00a2 -> LATIN SMALL LETTER O WITH ACUTE
    u'\xfa'  # 0x00a3 -> LATIN SMALL LETTER U WITH ACUTE
    u'\xf1'  # 0x00a4 -> LATIN SMALL LETTER N WITH TILDE
    u'\xd1'  # 0x00a5 -> LATIN CAPITAL LETTER N WITH TILDE
    u'\xaa'  # 0x00a6 -> FEMININE ORDINAL INDICATOR
    u'\xba'  # 0x00a7 -> MASCULINE ORDINAL INDICATOR
    u'\xbf'  # 0x00a8 -> INVERTED QUESTION MARK
    u'\u2310'  # 0x00a9 -> REVERSED NOT SIGN
    u'\xac'  # 0x00aa -> NOT SIGN
    u'\xbd'  # 0x00ab -> VULGAR FRACTION ONE HALF
    u'\xbc'  # 0x00ac -> VULGAR FRACTION ONE QUARTER
    u'\xa1'  # 0x00ad -> INVERTED EXCLAMATION MARK
    u'\xab'  # 0x00ae -> LEFT-POINTING DOUBLE ANGLE QUOTATION MARK
    u'\xbb'  # 0x00af -> RIGHT-POINTING DOUBLE ANGLE QUOTATION MARK
    u'\u2591'  # 0x00b0 -> LIGHT SHADE
    u'\u2592'  # 0x00b1 -> MEDIUM SHADE
    u'\u2593'  # 0x00b2 -> DARK SHADE
    u'\u2502'  # 0x00b3 -> BOX DRAWINGS LIGHT VERTICAL
    u'\u2524'  # 0x00b4 -> BOX DRAWINGS LIGHT VERTICAL AND LEFT
    u'\u2561'  # 0x00b5 -> BOX DRAWINGS VERTICAL SINGLE AND LEFT DOUBLE
    u'\u2562'  # 0x00b6 -> BOX DRAWINGS VERTICAL DOUBLE AND LEFT SINGLE
    u'\u2556'  # 0x00b7 -> BOX DRAWINGS DOWN DOUBLE AND LEFT SINGLE
    u'\u2555'  # 0x00b8 -> BOX DRAWINGS DOWN SINGLE AND LEFT DOUBLE
    u'\u2563'  # 0x00b9 -> BOX DRAWINGS DOUBLE VERTICAL AND LEFT
    u'\u2551'  # 0x00ba -> BOX DRAWINGS DOUBLE VERTICAL
    u'\u2557'  # 0x00bb -> BOX DRAWINGS DOUBLE DOWN AND LEFT
    u'\u255d'  # 0x00bc -> BOX DRAWINGS DOUBLE UP AND LEFT
    u'\u255c'  # 0x00bd -> BOX DRAWINGS UP DOUBLE AND LEFT SINGLE
    u'\u255b'  # 0x00be -> BOX DRAWINGS UP SINGLE AND LEFT DOUBLE
    u'\u2510'  # 0x00bf -> BOX DRAWINGS LIGHT DOWN AND LEFT
    u'\u2514'  # 0x00c0 -> BOX DRAWINGS LIGHT UP AND RIGHT
    u'\u2534'  # 0x00c1 -> BOX DRAWINGS LIGHT UP AND HORIZONTAL
    u'\u252c'  # 0x00c2 -> BOX DRAWINGS LIGHT DOWN AND HORIZONTAL
    u'\u251c'  # 0x00c3 -> BOX DRAWINGS LIGHT VERTICAL AND RIGHT
    u'\u2500'  # 0x00c4 -> BOX DRAWINGS LIGHT HORIZONTAL
    u'\u253c'  # 0x00c5 -> BOX DRAWINGS LIGHT VERTICAL AND HORIZONTAL
    u'\u255e'  # 0x00c6 -> BOX DRAWINGS VERTICAL SINGLE AND RIGHT DOUBLE
    u'\u255f'  # 0x00c7 -> BOX DRAWINGS VERTICAL DOUBLE AND RIGHT SINGLE
    u'\u255a'  # 0x00c8 -> BOX DRAWINGS DOUBLE UP AND RIGHT
    u'\u2554'  # 0x00c9 -> BOX DRAWINGS DOUBLE DOWN AND RIGHT
    u'\u2569'  # 0x00ca -> BOX DRAWINGS DOUBLE UP AND HORIZONTAL
    u'\u2566'  # 0x00cb -> BOX DRAWINGS DOUBLE DOWN AND HORIZONTAL
    u'\u2560'  # 0x00cc -> BOX DRAWINGS DOUBLE VERTICAL AND RIGHT
    u'\u2550'  # 0x00cd -> BOX DRAWINGS DOUBLE HORIZONTAL
    u'\u256c'  # 0x00ce -> BOX DRAWINGS DOUBLE VERTICAL AND HORIZONTAL
    u'\u2567'  # 0x00cf -> BOX DRAWINGS UP SINGLE AND HORIZONTAL DOUBLE
    u'\u2568'  # 0x00d0 -> BOX DRAWINGS UP DOUBLE AND HORIZONTAL SINGLE
    u'\u2564'  # 0x00d1 -> BOX DRAWINGS DOWN SINGLE AND HORIZONTAL DOUBLE
    u'\u2565'  # 0x00d2 -> BOX DRAWINGS DOWN DOUBLE AND HORIZONTAL SINGLE
    u'\u2559'  # 0x00d3 -> BOX DRAWINGS UP DOUBLE AND RIGHT SINGLE
    u'\u2558'  # 0x00d4 -> BOX DRAWINGS UP SINGLE AND RIGHT DOUBLE
    u'\u2552'  # 0x00d5 -> BOX DRAWINGS DOWN SINGLE AND RIGHT DOUBLE
    u'\u2553'  # 0x00d6 -> BOX DRAWINGS DOWN DOUBLE AND RIGHT SINGLE
    u'\u256b'  # 0x00d7 -> BOX DRAWINGS VERTICAL DOUBLE AND HORIZONTAL SINGLE
    u'\u256a'  # 0x00d8 -> BOX DRAWINGS VERTICAL SINGLE AND HORIZONTAL DOUBLE
    u'\u2518'  # 0x00d9 -> BOX DRAWINGS LIGHT UP AND LEFT
    u'\u250c'  # 0x00da -> BOX DRAWINGS LIGHT DOWN AND RIGHT
    u'\u2588'  # 0x00db -> FULL BLOCK
    u'\u2584'  # 0x00dc -> LOWER HALF BLOCK
    u'\u258c'  # 0x00dd -> LEFT HALF BLOCK
    u'\u2590'  # 0x00de -> RIGHT HALF BLOCK
    u'\u2580'  # 0x00df -> UPPER HALF BLOCK
    u'\u03b1'  # 0x00e0 -> GREEK SMALL LETTER ALPHA
    u'\xdf'  # 0x00e1 -> LATIN SMALL LETTER SHARP S
    u'\u0393'  # 0x00e2 -> GREEK CAPITAL LETTER GAMMA
    u'\u03c0'  # 0x00e3 -> GREEK SMALL LETTER PI
    u'\u03a3'  # 0x00e4 -> GREEK CAPITAL LETTER SIGMA
    u'\u03c3'  # 0x00e5 -> GREEK SMALL LETTER SIGMA
    u'\xb5'  # 0x00e6 -> MICRO SIGN
    u'\u03c4'  # 0x00e7 -> GREEK SMALL LETTER TAU
    u'\u03a6'  # 0x00e8 -> GREEK CAPITAL LETTER PHI
    u'\u0398'  # 0x00e9 -> GREEK CAPITAL LETTER THETA
    u'\u03a9'  # 0x00ea -> GREEK CAPITAL LETTER OMEGA
    u'\u03b4'  # 0x00eb -> GREEK SMALL LETTER DELTA
    u'\u221e'  # 0x00ec -> INFINITY
    u'\u03c6'  # 0x00ed -> GREEK SMALL LETTER PHI
    u'\u03b5'  # 0x00ee -> GREEK SMALL LETTER EPSILON
    u'\u2229'  # 0x00ef -> INTERSECTION
    u'\u2261'  # 0x00f0 -> IDENTICAL TO
    u'\xb1'  # 0x00f1 -> PLUS-MINUS SIGN
    u'\u2265'  # 0x00f2 -> GREATER-THAN OR EQUAL TO
    u'\u2264'  # 0x00f3 -> LESS-THAN OR EQUAL TO
    u'\u2320'  # 0x00f4 -> TOP HALF INTEGRAL
    u'\u2321'  # 0x00f5 -> BOTTOM HALF INTEGRAL
    u'\xf7'  # 0x00f6 -> DIVISION SIGN
    u'\u2248'  # 0x00f7 -> ALMOST EQUAL TO
    u'\xb0'  # 0x00f8 -> DEGREE SIGN
    u'\u2219'  # 0x00f9 -> BULLET OPERATOR
    u'\xb7'  # 0x00fa -> MIDDLE DOT
    u'\u221a'  # 0x00fb -> SQUARE ROOT
    u'\u207f'  # 0x00fc -> SUPERSCRIPT LATIN SMALL LETTER N
    u'\xb2'  # 0x00fd -> SUPERSCRIPT TWO
    u'\u25a0'  # 0x00fe -> BLACK SQUARE
    u'\xa0'  # 0x00ff -> NO-BREAK SPACE
)

# Encoding Map

ENCODING_MAP = {
    0x0000: 0x0000,  # NULL
    0x0001: 0x0001,  # START OF HEADING
    0x0002: 0x0002,  # START OF TEXT
    0x0003: 0x0003,  # END OF TEXT
    0x0004: 0x0004,  # END OF TRANSMISSION
    0x0005: 0x0005,  # ENQUIRY
    0x0006: 0x0006,  # ACKNOWLEDGE
    0x0007: 0x0007,  # BELL
    0x0008: 0x0008,  # BACKSPACE
    0x0009: 0x0009,  # HORIZONTAL TABULATION
    0x000a: 0x000a,  # LINE FEED
    0x000b: 0x000b,  # VERTICAL TABULATION
    0x000c: 0x000c,  # FORM FEED
    0x000d: 0x000d,  # CARRIAGE RETURN
    0x000e: 0x000e,  # SHIFT OUT
    0x000f: 0x000f,  # SHIFT IN
    0x0010: 0x0010,  # DATA LINK ESCAPE
    0x0011: 0x0011,  # DEVICE CONTROL ONE
    0x0012: 0x0012,  # DEVICE CONTROL TWO
    0x0013: 0x0013,  # DEVICE CONTROL THREE
    0x0014: 0x0014,  # DEVICE CONTROL FOUR
    0x0015: 0x0015,  # NEGATIVE ACKNOWLEDGE
    0x0016: 0x0016,  # SYNCHRONOUS IDLE
    0x0017: 0x0017,  # END OF TRANSMISSION BLOCK
    0x0018: 0x0018,  # CANCEL
    0x0019: 0x0019,  # END OF MEDIUM
    0x001a: 0x001a,  # SUBSTITUTE
    0x001b: 0x001b,  # ESCAPE
    0x001c: 0x001c,  # FILE SEPARATOR
    0x001d: 0x001d,  # GROUP SEPARATOR
    0x001e: 0x001e,  # RECORD SEPARATOR
    0x001f: 0x001f,  # UNIT SEPARATOR
    0x0020: 0x0020,  # SPACE
    0x0021: 0x0021,  # EXCLAMATION MARK
    0x0022: 0x0022,  # QUOTATION MARK
    0x0023: 0x0023,  # NUMBER SIGN
    0x0024: 0x0024,  # DOLLAR SIGN
    0x0025: 0x0025,  # PERCENT SIGN
    0x0026: 0x0026,  # AMPERSAND
    0x0027: 0x0027,  # APOSTROPHE
    0x0028: 0x0028,  # LEFT PARENTHESIS
    0x0029: 0x0029,  # RIGHT PARENTHESIS
    0x002a: 0x002a,  # ASTERISK
    0x002b: 0x002b,  # PLUS SIGN
    0x002c: 0x002c,  # COMMA
    0x002d: 0x002d,  # HYPHEN-MINUS
    0x002e: 0x002e,  # FULL STOP
    0x002f: 0x002f,  # SOLIDUS
    0x0030: 0x0030,  # DIGIT ZERO
    0x0031: 0x0031,  # DIGIT ONE
    0x0032: 0x0032,  # DIGIT TWO
    0x0033: 0x0033,  # DIGIT THREE
    0x0034: 0x0034,  # DIGIT FOUR
    0x0035: 0x0035,  # DIGIT FIVE
    0x0036: 0x0036,  # DIGIT SIX
    0x0037: 0x0037,  # DIGIT SEVEN
    0x0038: 0x0038,  # DIGIT EIGHT
    0x0039: 0x0039,  # DIGIT NINE
    0x003a: 0x003a,  # COLON
    0x003b: 0x003b,  # SEMICOLON
    0x003c: 0x003c,  # LESS-THAN SIGN
    0x003d: 0x003d,  # EQUALS SIGN
    0x003e: 0x003e,  # GREATER-THAN SIGN
    0x003f: 0x003f,  # QUESTION MARK
    0x0040: 0x0040,  # COMMERCIAL AT
    0x0041: 0x0041,  # LATIN CAPITAL LETTER A
    0x0042: 0x0042,  # LATIN CAPITAL LETTER B
    0x0043: 0x0043,  # LATIN CAPITAL LETTER C
    0x0044: 0x0044,  # LATIN CAPITAL LETTER D
    0x0045: 0x0045,  # LATIN CAPITAL LETTER E
    0x0046: 0x0046,  # LATIN CAPITAL LETTER F
    0x0047: 0x0047,  # LATIN CAPITAL LETTER G
    0x0048: 0x0048,  # LATIN CAPITAL LETTER H
    0x0049: 0x0049,  # LATIN CAPITAL LETTER I
    0x004a: 0x004a,  # LATIN CAPITAL LETTER J
    0x004b: 0x004b,  # LATIN CAPITAL LETTER K
    0x004c: 0x004c,  # LATIN CAPITAL LETTER L
    0x004d: 0x004d,  # LATIN CAPITAL LETTER M
    0x004e: 0x004e,  # LATIN CAPITAL LETTER N
    0x004f: 0x004f,  # LATIN CAPITAL LETTER O
    0x0050: 0x0050,  # LATIN CAPITAL LETTER P
    0x0051: 0x0051,  # LATIN CAPITAL LETTER Q
    0x0052: 0x0052,  # LATIN CAPITAL LETTER R
    0x0053: 0x0053,  # LATIN CAPITAL LETTER S
    0x0054: 0x0054,  # LATIN CAPITAL LETTER T
    0x0055: 0x0055,  # LATIN CAPITAL LETTER U
    0x0056: 0x0056,  # LATIN CAPITAL LETTER V
    0x0057: 0x0057,  # LATIN CAPITAL LETTER W
    0x0058: 0x0058,  # LATIN CAPITAL LETTER X
    0x0059: 0x0059,  # LATIN CAPITAL LETTER Y
    0x005a: 0x005a,  # LATIN CAPITAL LETTER Z
    0x005b: 0x005b,  # LEFT SQUARE BRACKET
    0x005c: 0x005c,  # REVERSE SOLIDUS
    0x005d: 0x005d,  # RIGHT SQUARE BRACKET
    0x005e: 0x005e,  # CIRCUMFLEX ACCENT
    0x005f: 0x005f,  # LOW LINE
    0x0060: 0x0060,  # GRAVE ACCENT
    0x0061: 0x0061,  # LATIN SMALL LETTER A
    0x0062: 0x0062,  # LATIN SMALL LETTER B
    0x0063: 0x0063,  # LATIN SMALL LETTER C
    0x0064: 0x0064,  # LATIN SMALL LETTER D
    0x0065: 0x0065,  # LATIN SMALL LETTER E
    0x0066: 0x0066,  # LATIN SMALL LETTER F
    0x0067: 0x0067,  # LATIN SMALL LETTER G
    0x0068: 0x0068,  # LATIN SMALL LETTER H
    0x0069: 0x0069,  # LATIN SMALL LETTER I
    0x006a: 0x006a,  # LATIN SMALL LETTER J
    0x006b: 0x006b,  # LATIN SMALL LETTER K
    0x006c: 0x006c,  # LATIN SMALL LETTER L
    0x006d: 0x006d,  # LATIN SMALL LETTER M
    0x006e: 0x006e,  # LATIN SMALL LETTER N
    0x006f: 0x006f,  # LATIN SMALL LETTER O
    0x0070: 0x0070,  # LATIN SMALL LETTER P
    0x0071: 0x0071,  # LATIN SMALL LETTER Q
    0x0072: 0x0072,  # LATIN SMALL LETTER R
    0x0073: 0x0073,  # LATIN SMALL LETTER S
    0x0074: 0x0074,  # LATIN SMALL LETTER T
    0x0075: 0x0075,  # LATIN SMALL LETTER U
    0x0076: 0x0076,  # LATIN SMALL LETTER V
    0x0077: 0x0077,  # LATIN SMALL LETTER W
    0x0078: 0x0078,  # LATIN SMALL LETTER X
    0x0079: 0x0079,  # LATIN SMALL LETTER Y
    0x007a: 0x007a,  # LATIN SMALL LETTER Z
    0x007b: 0x007b,  # LEFT CURLY BRACKET
    0x007c: 0x007c,  # VERTICAL LINE
    0x007d: 0x007d,  # RIGHT CURLY BRACKET
    0x007e: 0x007e,  # TILDE
    0x007f: 0x007f,  # DELETE
    0x00a0: 0x00ff,  # NO-BREAK SPACE
    0x00a1: 0x00ad,  # INVERTED EXCLAMATION MARK
    0x00a2: 0x009b,  # CENT SIGN
    0x00a3: 0x009c,  # POUND SIGN
    0x00a5: 0x009d,  # YEN SIGN
    0x00aa: 0x00a6,  # FEMININE ORDINAL INDICATOR
    0x00ab: 0x00ae,  # LEFT-POINTING DOUBLE ANGLE QUOTATION MARK
    0x00ac: 0x00aa,  # NOT SIGN
    0x00b0: 0x00f8,  # DEGREE SIGN
    0x00b1: 0x00f1,  # PLUS-MINUS SIGN
    0x00b2: 0x00fd,  # SUPERSCRIPT TWO
    0x00b5: 0x00e6,  # MICRO SIGN
    0x00b7: 0x00fa,  # MIDDLE DOT
    0x00ba: 0x00a7,  # MASCULINE ORDINAL INDICATOR
    0x00bb: 0x00af,  # RIGHT-POINTING DOUBLE ANGLE QUOTATION MARK
    0x00bc: 0x00ac,  # VULGAR FRACTION ONE QUARTER
    0x00bd: 0x00ab,  # VULGAR FRACTION ONE HALF
    0x00bf: 0x00a8,  # INVERTED QUESTION MARK
    0x00c4: 0x008e,  # LATIN CAPITAL LETTER A WITH DIAERESIS
    0x00c5: 0x008f,  # LATIN CAPITAL LETTER A WITH RING ABOVE
    0x00c6: 0x0092,  # LATIN CAPITAL LIGATURE AE
    0x00c7: 0x0080,  # LATIN CAPITAL LETTER C WITH CEDILLA
    0x00c9: 0x0090,  # LATIN CAPITAL LETTER E WITH ACUTE
    0x00d1: 0x00a5,  # LATIN CAPITAL LETTER N WITH TILDE
    0x00d6: 0x0099,  # LATIN CAPITAL LETTER O WITH DIAERESIS
    0x00dc: 0x009a,  # LATIN CAPITAL LETTER U WITH DIAERESIS
    0x00df: 0x00e1,  # LATIN SMALL LETTER SHARP S
    0x00e0: 0x0085,  # LATIN SMALL LETTER A WITH GRAVE
    0x00e1: 0x00a0,  # LATIN SMALL LETTER A WITH ACUTE
    0x00e2: 0x0083,  # LATIN SMALL LETTER A WITH CIRCUMFLEX
    0x00e4: 0x0084,  # LATIN SMALL LETTER A WITH DIAERESIS
    0x00e5: 0x0086,  # LATIN SMALL LETTER A WITH RING ABOVE
    0x00e6: 0x0091,  # LATIN SMALL LIGATURE AE
    0x00e7: 0x0087,  # LATIN SMALL LETTER C WITH CEDILLA
    0x00e8: 0x008a,  # LATIN SMALL LETTER E WITH GRAVE
    0x00e9: 0x0082,  # LATIN SMALL LETTER E WITH ACUTE
    0x00ea: 0x0088,  # LATIN SMALL LETTER E WITH CIRCUMFLEX
    0x00eb: 0x0089,  # LATIN SMALL LETTER E WITH DIAERESIS
    0x00ec: 0x008d,  # LATIN SMALL LETTER I WITH GRAVE
    0x00ed: 0x00a1,  # LATIN SMALL LETTER I WITH ACUTE
    0x00ee: 0x008c,  # LATIN SMALL LETTER I WITH CIRCUMFLEX
    0x00ef: 0x008b,  # LATIN SMALL LETTER I WITH DIAERESIS
    0x00f1: 0x00a4,  # LATIN SMALL LETTER N WITH TILDE
    0x00f2: 0x0095,  # LATIN SMALL LETTER O WITH GRAVE
    0x00f3: 0x00a2,  # LATIN SMALL LETTER O WITH ACUTE
    0x00f4: 0x0093,  # LATIN SMALL LETTER O WITH CIRCUMFLEX
    0x00f6: 0x0094,  # LATIN SMALL LETTER O WITH DIAERESIS
    0x00f7: 0x00f6,  # DIVISION SIGN
    0x00f9: 0x0097,  # LATIN SMALL LETTER U WITH GRAVE
    0x00fa: 0x00a3,  # LATIN SMALL LETTER U WITH ACUTE
    0x00fb: 0x0096,  # LATIN SMALL LETTER U WITH CIRCUMFLEX
    0x00fc: 0x0081,  # LATIN SMALL LETTER U WITH DIAERESIS
    0x00ff: 0x0098,  # LATIN SMALL LETTER Y WITH DIAERESIS
    0x0192: 0x009f,  # LATIN SMALL LETTER F WITH HOOK
    0x0393: 0x00e2,  # GREEK CAPITAL LETTER GAMMA
    0x0398: 0x00e9,  # GREEK CAPITAL LETTER THETA
    0x03a3: 0x00e4,  # GREEK CAPITAL LETTER SIGMA
    0x03a6: 0x00e8,  # GREEK CAPITAL LETTER PHI
    0x03a9: 0x00ea,  # GREEK CAPITAL LETTER OMEGA
    0x03b1: 0x00e0,  # GREEK SMALL LETTER ALPHA
    0x03b4: 0x00eb,  # GREEK SMALL LETTER DELTA
    0x03b5: 0x00ee,  # GREEK SMALL LETTER EPSILON
    0x03c0: 0x00e3,  # GREEK SMALL LETTER PI
    0x03c3: 0x00e5,  # GREEK SMALL LETTER SIGMA
    0x03c4: 0x00e7,  # GREEK SMALL LETTER TAU
    0x03c6: 0x00ed,  # GREEK SMALL LETTER PHI
    0x207f: 0x00fc,  # SUPERSCRIPT LATIN SMALL LETTER N
    0x20a7: 0x009e,  # PESETA SIGN
    0x2219: 0x00f9,  # BULLET OPERATOR
    0x221a: 0x00fb,  # SQUARE ROOT
    0x221e: 0x00ec,  # INFINITY
    0x2229: 0x00ef,  # INTERSECTION
    0x2248: 0x00f7,  # ALMOST EQUAL TO
    0x2261: 0x00f0,  # IDENTICAL TO
    0x2264: 0x00f3,  # LESS-THAN OR EQUAL TO
    0x2265: 0x00f2,  # GREATER-THAN OR EQUAL TO
    0x2310: 0x00a9,  # REVERSED NOT SIGN
    0x2320: 0x00f4,  # TOP HALF INTEGRAL
    0x2321: 0x00f5,  # BOTTOM HALF INTEGRAL
    0x2500: 0x00c4,  # BOX DRAWINGS LIGHT HORIZONTAL
    0x2502: 0x00b3,  # BOX DRAWINGS LIGHT VERTICAL
    0x250c: 0x00da,  # BOX DRAWINGS LIGHT DOWN AND RIGHT
    0x2510: 0x00bf,  # BOX DRAWINGS LIGHT DOWN AND LEFT
    0x2514: 0x00c0,  # BOX DRAWINGS LIGHT UP AND RIGHT
    0x2518: 0x00d9,  # BOX DRAWINGS LIGHT UP AND LEFT
    0x251c: 0x00c3,  # BOX DRAWINGS LIGHT VERTICAL AND RIGHT
    0x2524: 0x00b4,  # BOX DRAWINGS LIGHT VERTICAL AND LEFT
    0x252c: 0x00c2,  # BOX DRAWINGS LIGHT DOWN AND HORIZONTAL
    0x2534: 0x00c1,  # BOX DRAWINGS LIGHT UP AND HORIZONTAL
    0x253c: 0x00c5,  # BOX DRAWINGS LIGHT VERTICAL AND HORIZONTAL
    0x2550: 0x00cd,  # BOX DRAWINGS DOUBLE HORIZONTAL
    0x2551: 0x00ba,  # BOX DRAWINGS DOUBLE VERTICAL
    0x2552: 0x00d5,  # BOX DRAWINGS DOWN SINGLE AND RIGHT DOUBLE
    0x2553: 0x00d6,  # BOX DRAWINGS DOWN DOUBLE AND RIGHT SINGLE
    0x2554: 0x00c9,  # BOX DRAWINGS DOUBLE DOWN AND RIGHT
    0x2555: 0x00b8,  # BOX DRAWINGS DOWN SINGLE AND LEFT DOUBLE
    0x2556: 0x00b7,  # BOX DRAWINGS DOWN DOUBLE AND LEFT SINGLE
    0x2557: 0x00bb,  # BOX DRAWINGS DOUBLE DOWN AND LEFT
    0x2558: 0x00d4,  # BOX DRAWINGS UP SINGLE AND RIGHT DOUBLE
    0x2559: 0x00d3,  # BOX DRAWINGS UP DOUBLE AND RIGHT SINGLE
    0x255a: 0x00c8,  # BOX DRAWINGS DOUBLE UP AND RIGHT
    0x255b: 0x00be,  # BOX DRAWINGS UP SINGLE AND LEFT DOUBLE
    0x255c: 0x00bd,  # BOX DRAWINGS UP DOUBLE AND LEFT SINGLE
    0x255d: 0x00bc,  # BOX DRAWINGS DOUBLE UP AND LEFT
    0x255e: 0x00c6,  # BOX DRAWINGS VERTICAL SINGLE AND RIGHT DOUBLE
    0x255f: 0x00c7,  # BOX DRAWINGS VERTICAL DOUBLE AND RIGHT SINGLE
    0x2560: 0x00cc,  # BOX DRAWINGS DOUBLE VERTICAL AND RIGHT
    0x2561: 0x00b5,  # BOX DRAWINGS VERTICAL SINGLE AND LEFT DOUBLE
    0x2562: 0x00b6,  # BOX DRAWINGS VERTICAL DOUBLE AND LEFT SINGLE
    0x2563: 0x00b9,  # BOX DRAWINGS DOUBLE VERTICAL AND LEFT
    0x2564: 0x00d1,  # BOX DRAWINGS DOWN SINGLE AND HORIZONTAL DOUBLE
    0x2565: 0x00d2,  # BOX DRAWINGS DOWN DOUBLE AND HORIZONTAL SINGLE
    0x2566: 0x00cb,  # BOX DRAWINGS DOUBLE DOWN AND HORIZONTAL
    0x2567: 0x00cf,  # BOX DRAWINGS UP SINGLE AND HORIZONTAL DOUBLE
    0x2568: 0x00d0,  # BOX DRAWINGS UP DOUBLE AND HORIZONTAL SINGLE
    0x2569: 0x00ca,  # BOX DRAWINGS DOUBLE UP AND HORIZONTAL
    0x256a: 0x00d8,  # BOX DRAWINGS VERTICAL SINGLE AND HORIZONTAL DOUBLE
    0x256b: 0x00d7,  # BOX DRAWINGS VERTICAL DOUBLE AND HORIZONTAL SINGLE
    0x256c: 0x00ce,  # BOX DRAWINGS DOUBLE VERTICAL AND HORIZONTAL
    0x2580: 0x00df,  # UPPER HALF BLOCK
    0x2584: 0x00dc,  # LOWER HALF BLOCK
    0x2588: 0x00db,  # FULL BLOCK
    0x258c: 0x00dd,  # LEFT HALF BLOCK
    0x2590: 0x00de,  # RIGHT HALF BLOCK
    0x2591: 0x00b0,  # LIGHT SHADE
    0x2592: 0x00b1,  # MEDIUM SHADE
    0x2593: 0x00b2,  # DARK SHADE
    0x25a0: 0x00fe,  # BLACK SQUARE
}



#!/usr/bin/env python2.7
""" web server for x/84. """
import threading
import traceback
import logging
import web
import sys
import os


class Favicon(object):

    """ Dummy class for preventing 404 of ``/favicon.ico`` """

    def GET(self):
        """ GET request callback (does nothing). """
        pass


def _get_fp(section_key, optional=False):
    """ Return filepath of [web] option by ``section_key``. """
    from x84.bbs import get_ini
    value = get_ini(section='web', key=section_key) or None
    if value:
        value = os.path.expanduser(value)
    elif optional:
        return None
    assert value is not None and os.path.isfile(value), (
        'Configuration section [web], key `{section_key}`, '
        'must {optional_exist}identify a path to an '
        'SSL {section_key} file. '
        '(value is {value})'.format(
            section_key=section_key,
            optional_exist=not optional and 'exist and ' or '',
            value=value))
    return value


def get_urls_funcs(web_modules):
    """ Get url function mapping for the given web modules. """
    log = logging.getLogger(__name__)

    # list of url's to route to each module api; defaults to route /favicon.ico
    # to a non-op to avoid 404 errors.
    # See: http://webpy.org/docs/0.3/api#web.application
    urls = ('/favicon.ico', 'favicon')
    funcs = globals()
    funcs['favicon'] = Favicon

    log.debug('add url {0} => {1}'.format(
        '/favicon.ico', Favicon.__name__))

    for mod in web_modules:
        module = None

        # first, check in system PATH (includes SCRIPT_PATH)
        try:
            module = __import__('webmodules.{0}'.format(mod),
                                fromlist=('webmodules',))
        except ImportError:
            # failed to import, check in x84's path, raise naturally
            module = __import__('x84.webmodules.{0}'.format(mod),
                                fromlist=('x84.webmodules',))

        api = module.web_module()

        for key in api['funcs']:
            funcs[key] = api['funcs'][key]

        # use zip to transform (1,2,3,4,5,6,7,8) =>
        # [(1, 2), (2, 3), (3, 4), (4, 5), (5, 6), (6, 7), (7, 8)]
        # then, use slice to 'step 2' =>
        # [(1, 2), (3, 4), (5, 6), (7, 8)]
        for (url, f_key) in zip(api['urls'], api['urls'][1:])[::2]:
            if f_key not in funcs:
                log.error('module {module} provided url {url_tuple} without '
                          'matching function (available: {f_avail})'
                          .format(module=module,
                                  url_tuple=(url, f_key,),
                                  f_avail=funcs.keys()))
            else:
                log.debug('add url {0} => {1}'.format(
                    url, funcs[f_key].__name__))
        urls += api['urls']
    return urls, funcs


def server(urls, funcs):
    """ Main server thread for running the web server """
    from x84.bbs import get_ini
    from web.wsgiserver import CherryPyWSGIServer
    from web.wsgiserver.ssl_pyopenssl import pyOpenSSLAdapter
    from OpenSSL import SSL

    log = logging.getLogger(__name__)

    cert, key, chain = (_get_fp('cert'),
                        _get_fp('key'),
                        _get_fp('chain', optional=True))

    addr = get_ini(section='web',
                   key='addr'
                   ) or '0.0.0.0'

    port = get_ini(section='web',
                   key='port',
                   getter='getint'
                   ) or 8443

    # List of ciphers made available, composed by haliphax without reference,
    # but apparently to prevent POODLE? This stuff is hard -- the best source
    # would probably be to compare by cloudflare's latest sslconfig file:
    #
    #   https://github.com/cloudflare/sslconfig/blob/master/conf
    #
    cipher_list = (get_ini(section='web', key='cipher_list')
                   or ':'.join((
                       'ECDH+AESGCM',
                       'ECDH+AES256',
                       'ECDH+AES128',
                       'ECDH+3DES',
                       'DH+AESGCM',
                       'DH+AES256',
                       'DH+AES',
                       'DH+3DES',
                       'RSA+AESGCM',
                       'RSA+AES',
                       'RSA+3DES',
                       '!aNULL',
                       '!MD5',
                       '!DSS',
                   )))

    CherryPyWSGIServer.ssl_adapter = pyOpenSSLAdapter(cert, key, chain)
    CherryPyWSGIServer.ssl_adapter.context = SSL.Context(SSL.SSLv23_METHOD)
    CherryPyWSGIServer.ssl_adapter.context.set_options(SSL.OP_NO_SSLv3)

    try:
        CherryPyWSGIServer.ssl_adapter.context.use_certificate_file(cert)
    except Exception:
        # wrap exception to contain filepath to 'cert' file, which will
        # hopefully help the user better understand what otherwise be very
        # obscure.
        error = ''.join(
            traceback.format_exception_only(
                sys.exc_info()[0],
                sys.exc_info()[1])).rstrip()
        raise ValueError('Exception loading ssl certificate file {0!r}: '
                         '{1}'.format(cert, error))

    try:
        CherryPyWSGIServer.ssl_adapter.context.use_privatekey_file(key)
    except Exception:
        # also wrap exception to contain filepath to 'key' file.
        error = ''.join(
            traceback.format_exception_only(
                sys.exc_info()[0],
                sys.exc_info()[1])).rstrip()
        raise ValueError('Exception loading ssl key file {0!r}: '
                         '{1}'.format(key, error))

    if chain is not None:
        (CherryPyWSGIServer.ssl_adapter.context
         .use_certificate_chain_file(chain))

    CherryPyWSGIServer.ssl_adapter.context.set_cipher_list(cipher_list)

    app = web.application(urls, funcs)

    web.config.debug = False

    log.info('https listening on {addr}:{port}/tcp'
             .format(addr=addr, port=port))

    # Runs CherryPy WSGI server hosting WSGI app.wsgifunc().
    web.httpserver.runsimple(app.wsgifunc(), (addr, port))  # blocking


def main(background_daemon=True):
    """
    Entry point to configure and begin web server.

    Called by x84/engine.py, function main() as unmanaged thread.

    :param bool background_daemon: When True (default), this function returns
       and web modules are served in an unmanaged, background (daemon) thread.
       Otherwise, function call to ``main()`` is blocking.
    :rtype: None
    """
    from x84.bbs import get_ini

    log = logging.getLogger(__name__)

    SCRIPT_PATH = get_ini(section='system', key='scriptpath')

    # ensure the SCRIPT_PATH is in os environment PATH for module lookup.
    sys.path.insert(0, os.path.expanduser(SCRIPT_PATH))

    web_modules = get_ini(section='web', key='modules', split=True)

    if not web_modules:
        log.error("web server enabled, but no `modules' "
                  "defined in section [web]")
        return

    log.debug(u'Ready web modules: {0}'.format(web_modules))
    urls, funcs = get_urls_funcs(web_modules)

    if background_daemon:
        t = threading.Thread(target=server, args=(urls, funcs,))
        t.daemon = True
        t.start()
    else:
        server(urls=urls, funcs=funcs)


if __name__ == '__main__':
    # load only the webserver module when executing this script directly.
    #
    # as we are running outside of the 'engine' context, it is necessary
    # for us to initialize the .ini configuration scheme so that the list
    # of web modules and ssl options may be gathered.
    import x84.bbs.ini
    import x84.cmdline
    x84.bbs.ini.init(*x84.cmdline.parse_args())

    # do not execute webserver as a background thread.
    main(background_daemon=False)



""" Lightbar package for x/84. """

# local imports
from x84.bbs.ansiwin import AnsiWindow
from x84.bbs.session import getterminal, getch
from x84.bbs.output import decode_pipe, echo


#: default command-key mapping.
NETHACK_KEYSET = {'home': [u'y', '0'],
                  'end': [u'n', 'G'],
                  'pgup': [u'h', u'b'],
                  'pgdown': [u'l', u'f'],
                  'up': [u'k'],
                  'down': [u'j'],
                  'enter': [u'\r'],
                  'exit': [u'q', u'Q', unichr(27), ],
                  }


class Lightbar(AnsiWindow):

    """
    This Windowing class offers a classic 'lightbar' interface.

    Instantiate with yloc, xloc, height, and width, then call the update method
    with a list of unicode strings. send keycodes to process_keystroke () to
    interactive with the 'lightbar'.
    """

    # pylint: disable=R0902
    #         Too many instance attributes (15/7)
    # pylint: disable=R0904
    #         Too many public methods (29/20)
    def __init__(self, *args, **kwargs):
        """
        Class initializer.

        Initialize a lightbar of height, width, y and x, and position.

        :param int width: width of window.
        :param int height: height of window.
        :param int yloc: y-location of window.
        :param int xloc: x-location of window.
        :param dict colors: color theme, only key value of ``highlight``
                            is used.
        :param dict glyphs: bordering window character glyphs.
        :param dict keyset: command keys, global ``NETHACK_KEYSET`` is
                            used by default, augmented by application
                            keys such as home, end, pgup, etc.
        :param list content: Lightbar content as list of tuples, an empty list
                             is used by default.  Tuples must be in form of
                             ``(key, str)``.  ``key`` may have any suitable
                             significance for the caller.  ``str``, however,
                             must be of a unicode terminal sequence.
        """
        self._selected = False
        self._quit = False
        self._vitem_idx = self._vitem_shift = -1
        self.content = kwargs.pop('content', list())

        pos = kwargs.pop('position', (0, 0)) or (0, 0)

        self.init_keystrokes(
            keyset=kwargs.pop('keyset', NETHACK_KEYSET.copy()))

        AnsiWindow.__init__(self, *args, **kwargs)
        self.position = pos

    def init_theme(self, colors=None, glyphs=None):
        """ Set color and bordering glyphs theme. """
        colors = colors or {'highlight': getterminal().reverse_yellow}
        glyphs = glyphs or {'strip': u' $'}
        AnsiWindow.init_theme(self, colors, glyphs)

    def init_keystrokes(self, keyset):
        """ Sets keyboard keys for various editing keystrokes. """
        term = getterminal()
        self.keyset = keyset
        self.keyset['home'].append(term.KEY_HOME)
        self.keyset['end'].append(term.KEY_END)
        self.keyset['pgup'].append(term.KEY_PGUP)
        self.keyset['pgdown'].append(term.KEY_PGDOWN)
        self.keyset['up'].append(term.KEY_UP)
        self.keyset['down'].append(term.KEY_DOWN)
        self.keyset['enter'].append(term.KEY_ENTER)
        self.keyset['exit'].append(term.KEY_ESCAPE)

    def update(self, keyed_uchars=None):
        """ Replace content with with sequence of (key, str).

        ``key`` may have any suitable significance for the caller.  ``str``,
        however, must be of a unicode terminal sequence.
        """
        if keyed_uchars is None:
            keyed_uchars = (None, u'',)
        self.content = list(keyed_uchars)
        self.position = (self.vitem_idx, self.vitem_shift)

    def refresh_row(self, row):
        """ Return string sequence suitable for refreshing current selection.

        Return unicode byte sequence suitable for moving to location ypos of
        window-relative row, and displaying any valid entry there, or using
        glyphs['erase'] if out of bounds. Strings are ansi color safe, and
        will be trimmed using glyphs['strip'] if their displayed width is
        wider than window.
        """
        term = getterminal()

        pos = self.pos(self.ypadding + row, self.xpadding)
        entry = self.vitem_shift + row
        if entry >= len(self.content):
            # out-of-bounds;
            disp_erase = self.glyphs.get('erase', u' ') * self.visible_width
            return u''.join((pos, disp_erase,))

        def fit_row(ucs):
            """ Strip a unicode row to fit window boundry, if necessary """
            column = self.visible_width - 1
            wrapped = term.wrap(ucs, column)
            if len(wrapped) > 1:
                marker = self.glyphs.get('strip', u' $')
                marker_column = self.visible_width - len(marker)
                wrapped = term.wrap(ucs, marker_column)
                ucs = term.ljust(wrapped[0].rstrip(), marker_column) + marker
                return ucs
            return term.ljust(ucs, column)

        # allow ucs data with '\r\n', to accomidate soft and hardbreaks; just
        # don't display them, really wrecks up cusor positioning.
        ucs = self.content[entry][1].strip(u'\r\n')

        # highlighted entry; strip of ansi sequences, use color 'highlight'
        # trim and append '$ ' if it cannot fit,
        if entry == self.index:
            ucs = term.strip_seqs(ucs)
            if term.length(ucs) > self.visible_width:
                ucs = fit_row(ucs)
            return u''.join((pos,
                             term.normal,
                             self.colors.get('highlight', u''),
                             self.align(ucs),
                             term.normal,))
        # unselected entry; retain ansi sequences, decode any pipe characters,
        # trim and append '$ ' if it cannot fit
        ucs = decode_pipe(ucs)
        if term.length(ucs) > self.visible_width:
            ucs = fit_row(ucs)
        return u''.join((pos,
                         self.colors.get('lowlight', u''),
                         self.align(ucs),
                         term.normal,))

    def fixate(self):
        """ Return string sequence suitable for "fixating" cursor position. """
        return self.pos(self.ypadding + self.vitem_idx,
                        self.xpadding + self.visible_width)

    def refresh(self):
        """ Return string sequence suitable for refreshing lightbar. """
        return u''.join(self.refresh_row(ypos) for ypos in
                        range(max(self.visible_bottom, self.visible_height)))

    def refresh_quick(self):
        """ Redraw only the 'dirty' portions after a 'move' has occurred. """
        if self.moved:
            if (self._vitem_lastshift != self.vitem_shift):
                # page shift, refresh entire page
                return self.refresh()
            if self._vitem_lastidx != self.vitem_idx:
                # unhighlight last selection, highlight new
                return (self.refresh_row(self._vitem_lastidx)
                        + self.refresh_row(self.vitem_idx))
            else:
                # just highlight new ..
                return (self.refresh_row(self.vitem_idx))
        return u''

    def process_keystroke(self, key):
        """
        Process the keystroke and return string to refresh.

        :param blessed.keyboard.Keystroke keystroke: input from ``inkey()``.
        :rtype: str
        :returns: string sequence suitable for refresh.
        """
        self._moved = False
        self._selected = False
        self._vitem_lastidx = self.vitem_idx
        self._vitem_lastshift = self.vitem_shift
        rstr = u''
        keystroke = hasattr(key, 'code') and key.code or key
        if keystroke in self.keyset['home']:
            rstr = self.move_home()
        elif keystroke in self.keyset['end']:
            rstr = self.move_end()
        elif keystroke in self.keyset['pgup']:
            rstr = self.move_pageup()
        elif keystroke in self.keyset['pgdown']:
            rstr = self.move_pagedown()
        elif keystroke in self.keyset['up']:
            rstr = self.move_up()
        elif keystroke in self.keyset['down']:
            rstr = self.move_down()
        elif keystroke in self.keyset['enter']:
            self.selected = True
        elif keystroke in self.keyset['exit']:
            self._quit = True
        return rstr

    def read(self):
        """
        Reads input until the ENTER or ESCAPE key is pressed (Blocking).

        Returns selection content, or None when canceled.
        """
        self._selected = False
        self._quit = False
        echo(self.refresh())
        while not (self.selected or self.quit):
            echo(self.process_keystroke(getch()))
        if self.quit:
            return None
        return self.selection[0]

    @property
    def quit(self):
        """ Whether a 'quit' character has been handled, such as escape. """
        return self._quit

    @property
    def index(self):
        """ Selected index of self.content. """
        return self.vitem_shift + self.vitem_idx

    @property
    def at_bottom(self):
        """ Whether current selection is pointed at final entry. """
        return self.index == len(self.content) - 1

    @property
    def at_top(self):
        """ Whether current selection is pointed at the first entry. """
        return self.index == 0

    @property
    def selection(self):
        """ Selected content of self.content by index. """
        return (self.content[self.index]
                if self.index >= 0 and self.index < len(self.content)
                else (None, None))

    @property
    def selected(self):
        """  Whether carriage return was detected by process_keystroke. """
        return self._selected

    @selected.setter
    def selected(self, value):
        # pylint: disable=C0111
        #         Missing docstring
        # this setter should only be used to reset to 'False' for recycling
        assert isinstance(value, bool)
        self._selected = value

    @property
    def last_index(self):
        """ Previously selected index of self.content. """
        return self._vitem_lastshift + self._vitem_lastidx

    @property
    def position(self):
        """
        Tuple pair (row, page).

        ``row`` is the index from top of window, and 'page' is number of page
        items scrolled.
        """
        return (self.vitem_idx, self.vitem_shift)

    @position.setter
    def position(self, pos_tuple):
        # pylint: disable=C0111
        #         Missing docstring
        self.vitem_idx, self.vitem_shift = pos_tuple
        self._chk_bounds()

    @property
    def visible_content(self):
        """ Returns visible content only. """
        _, shift = self.position
        return self.content[shift:shift + self.visible_height]

    @property
    def visible_bottom(self):
        """ Visible bottom-most item of lightbar. """
        if self.vitem_shift + (self.visible_height - 1) > len(self.content):
            return len(self.content)
        else:
            return self.visible_height

    @property
    def vitem_idx(self):
        """
        Relative visible item index within view.

        Index of selected item relative by index to only the length of the list
        that is visible, without accounting for scrolled content.
        """
        return self._vitem_idx

    @vitem_idx.setter
    def vitem_idx(self, value):
        # pylint: disable=C0111
        #         Missing docstring
        if self._vitem_idx != value:
            self._vitem_lastidx = self._vitem_idx
            self._moved = True
        self._vitem_idx = value

    @property
    def vitem_shift(self):
        """
        Index of top-most item in viewable window, non-zero when scrolled.

        This value effectively represents the number of items not in view
        due to paging.
        """
        return self._vitem_shift

    @vitem_shift.setter
    def vitem_shift(self, value):
        # pylint: disable=C0111
        #         Missing docstring
        if self._vitem_shift != value:
            self._vitem_lastshift = self._vitem_shift
            self._moved = True
        self._vitem_shift = value

    def _chk_bounds(self):
        """ Shift pages and selection until selection is within bounds. """
        # if selected item is out of range of new list, then scroll to last
        # page, and move selection to end of screen,
        if self.vitem_shift and (self.index + 1) > len(self.content):
            self.vitem_shift = len(self.content) - self.visible_height + 1
            self.vitem_idx = self.visible_height - 2

        # if we are a shifted window, shift 1 line up while keeping our
        # lightbar position until the bottom-most item is within visable range.
        while (self.vitem_shift and self.vitem_shift + self.visible_height - 1
                > len(self.content)):
            self.vitem_shift -= 1
            self.vitem_idx += 1

        # When a window is not shiftable, ensure selection is less than
        # total items. (truncate to last item)
        while self.vitem_idx > 0 and self.index >= len(self.content):
            self.vitem_idx -= 1

    def move_down(self):
        """ Move selection down one row, return string suitable for refresh. """
        if self.at_bottom:
            # bounds check
            return u''
        if self.vitem_idx + 1 < self.visible_bottom:
            # move down 1 row
            self.vitem_idx += 1
        elif self.vitem_idx < len(self.content):
            # scroll down 1 row
            self.vitem_shift += 1
        return self.refresh_quick()

    def goto(self, index):
        """ Move selection to given index. """
        assert index >= 0 and index < len(self.content)
        row, shift = self.position
        while (row + shift) < index:
            if row < (self.visible_height - 1):
                row += 1
            else:
                shift += 1
        while (row + shift) > index:
            if row > 0:
                row -= 1
            else:
                shift -= 1
        self.position = row, shift
        return self.refresh_quick()

    def move_up(self):
        """ Move selection up one row, return string suitable for refresh. """
        if self.at_top:
            # bounds check
            return u''
        elif self.vitem_idx >= 1:
            # move up 1 row
            self.vitem_idx -= 1
        elif self.vitem_shift > 0:
            # scroll up 1 row
            self.vitem_shift -= 1
        return self.refresh_quick()

    def move_pagedown(self):
        """ Move selection down one page, return string suitable for refresh. """
        if len(self.content) < self.visible_height:
            # move to last entry
            if self.vitem_idx == len(self.content) - 1:
                return u''  # already at end
            self.vitem_idx = len(self.content) - 1
        elif (self.vitem_shift + self.visible_height
                < (len(self.content) - self.visible_height)):
            # previous page
            self.vitem_shift = self.vitem_shift + self.visible_height
        elif self.vitem_shift != len(self.content) - self.visible_height:
            # shift window to last page
            self.vitem_shift = len(self.content) - self.visible_height
        else:
            # already at last page, goto end
            return self.move_end()
        return self.refresh_quick()

    def move_pageup(self):
        """ Move selection up one page, return string suitable for refresh. """
        if len(self.content) < self.visible_height - 1:
            self.vitem_idx = 0
        if self.vitem_shift - self.visible_height > 0:
            # previous page
            self.vitem_shift = self.vitem_shift - self.visible_height
        elif self.vitem_shift > 0:
            # shift window to first page
            self.vitem_shift = 0
        else:
            # already at first page, goto home
            return self.move_home()
        return self.refresh_quick()

    def move_home(self):
        """ Move selection to first row, return string suitable for refresh. """
        if (0, 0) == (self.vitem_idx, self.vitem_shift):
            return u''  # already at home
        self.vitem_idx = 0
        self.vitem_shift = 0
        return self.refresh_quick()

    def move_end(self):
        """ Move selection to final row, return string suitable for refresh. """
        if len(self.content) < self.visible_height:
            if self.vitem_idx == len(self.content) - 1:
                return u''  # already at end
            self.vitem_idx = len(self.content) - 1
        else:
            self.vitem_shift = len(self.content) - self.visible_height
            self.vitem_idx = self.visible_height - 1
        return self.refresh_quick()



"""
Door package for x/84.

This implements the concept of "Doors", popular for DOS BBS software.

It also supports executing external Unix paths. See wikipedia article
for details: http://en.wikipedia.org/wiki/BBS_door
"""

# std imports
import logging
import select
import codecs
import struct
import time
import sys
import os
import re

# local imports
from x84.bbs.session import getsession, getterminal
from x84.bbs.output import echo
from x84.bbs.ini import get_ini
from x84.bbs.userbase import list_users


class Dropfile(object):

    """
    Dropfile export class.

    From http://en.wikipedia.org/wiki/BBS_door

    > the 1990s on, most BBS software had the capability to "drop to" doors.
    > Several standards were developed for passing connection and user
    > information to doors; this was usually done with "dropfiles", small
    > binary or text files dropped into known locations in the BBS's file
    > system.
    """

    #: Dropfile type constants
    (DOORSYS, DOOR32, CALLINFOBBS, DORINFO) = range(4)

    def __init__(self, filetype=None, node=None):
        """
        Class initializer.

        :param int filetype: dropfile type. One of ``Dropfile.DOORSYS``,
                             ``Dropfile.DOOR32``, ``Dropfile.CALLINFOBBS``,
                             or ``Dropfile.DORINFO``.
        :param int node: A node number specified by caller; for some DOS
                         doors, this is a very specific and limited number
                         bounded and lock-acquired per-door by sesame.py.
                         For others, it is inconsequential, in which case
                         the session's system-wide node number is used.
        """
        self._filetype = filetype
        self._node = node

    def save(self, folder):
        """ Save dropfile to destination ``folder``. """
        f_path = os.path.join(folder, self.filename)
        with codecs.open(f_path, 'w', 'ascii', 'replace') as out_p:
            out_p.write(self.__str__())

    @property
    def node(self):
        """ User's node number. """
        return self._node or getsession().node

    @property
    def location(self):
        """ User location. """
        return getsession().user.location

    @property
    def fullname(self):
        """ User fullname. Returns ``<handle> <handle>``. """
        return '%s %s' % (
            getsession().user.handle,
            getsession().user.handle,)

    @property
    def securitylevel(self):
        """ User security level. Always 30, or 100 for sysop. """
        return 100 if getsession().user.is_sysop else 30

    @property
    def numcalls(self):
        """ Number of calls by user. """
        return getsession().user.calls

    @property
    def lastcall_date(self):
        """ Date of last call (format is ``%m/%d/%y``). """
        return time.strftime(
            '%m/%d/%y', time.localtime(getsession().user.lastcall))

    @property
    def lastcall_time(self):
        """ Time of last call (format is ``%H:%M``). """
        return time.strftime(
            '%H:%M', time.localtime(getsession().user.lastcall))

    @property
    def time_used(self):
        """ Time used (session duration) in seconds. """
        return int(time.time() - getsession().connect_time)

    @property
    def remaining_secs(self):
        """ Remaining seconds (always returns ``15360``). """
        return 256 * 60

    @property
    def remaining_mins(self):
        """ Remaining minutes (always returns ``256``). """
        return 256

    @property
    def comport(self):
        """ Com port (always returns ``COM1``). """
        return 'COM1'

    @property
    def comspeed(self):
        """ Com speed (always returns ``57600``). """
        return 57600

    @property
    def comtype(self):
        """ Com type (always returns ``0``). """
        return 0  # Line 1 : Comm type (0=local, 1=serial, 2=telnet)

    @property
    def comhandle(self):
        """ Com handle (always returns ``0``). """
        return 0  # Line 2 : Comm or socket handle

    @property
    def parity(self):
        """ Data parity. """
        return 8

    @property
    def password(self):
        """ Password of user. """
        return '<encrypted>'

    @property
    def pageheight(self):
        """ Terminal height. """
        return getterminal().height

    @property
    def systemname(self):
        """ BBS System name. """
        return get_ini('system', 'software') or 'x/84'

    @property
    def xferprotocol(self):
        """ preferred transfer protocol. """
        return 'X'  # x-modem for now, we don't have any xfer code/prefs

    @property
    def usernum(self):
        """ User record number. """
        try:
            return list_users().index(getsession().user.handle)
        except ValueError:
            return 999

    @property
    def sysopname(self):
        """ name of sysop. """
        return get_ini('system', 'sysop') or u''

    @property
    def alias(self):
        """ current session's handle. """
        return getsession().user.handle

    @property
    def filename(self):
        """ Filename of given dropfile. """
        if self._filetype == self.DOORSYS:
            return 'DOOR.SYS'
        elif self._filetype == self.DOOR32:
            return 'DOOR32.SYS'
        elif self._filetype == self.CALLINFOBBS:
            return 'CALLINFO.BBS'
        elif self._filetype == self.DORINFO:
            # n in DORINFO<n>.DEF is 1-9,0,a-z
            if self.node == 10:
                nodeid = '0'
            elif self.node < 10:
                nodeid = str(self.node)
            else:
                nodeid = chr(ord('a') + (self.node - 11))
                assert ord(nodeid) <= ord('z')
            return 'DORINFO{0}.DEF'.format(nodeid).upper()
        else:
            raise ValueError('filetype is unknown: {0}'.format(self._filetype))

    def __str__(self):
        """ Return dropfile content. """
        method = {
            self.DOORSYS: self._get_doorsys,
            self.DOOR32: self._get_door32,
            self.CALLINFOBBS: self._get_callinfo,
            self.DORINFO: self._get_dorinfo,
        }.get(self._filetype, None)
        if method is None:
            raise ValueError('unknown dropfile filetype: {self._filetype}'
                             .format(self=self))
        return method()

    def _get_doorsys(self):
        """ Return door.sys-formatted dropfile content. """
        return (u'{self.comport}:\r\n'
                u'{self.comspeed}\r\n'
                u'{self.parity}\r\n'
                u'{self.node}\r\n'
                u'{self.comspeed}\r\n'
                u'Y\r\n'                     # screen?
                u'Y\r\n'                     # printer?
                u'Y\r\n'                     # pager alarm?
                u'Y\r\n'                     # caller alartm?
                u'{self.fullname}\r\n'
                u'{self.location}\r\n'
                u'123-456-7890\r\n'          # phone number1
                u'123-456-7890\r\n'          # phone number2
                u'{self.password}\r\n'
                u'{self.securitylevel}\r\n'
                u'{self.numcalls}\r\n'
                u'{self.lastcall_date}\r\n'
                u'{self.remaining_secs}\r\n'
                u'{self.remaining_mins}\r\n'
                u'GR\r\n'                    # graphics mode
                u'{self.pageheight}\r\n'
                u'N\r\n'                     # expert mode?
                u'1,2,3,4,5,6,7\r\n'         # conferences
                u'1\r\n'                     # conf. sel, exp. date
                u'01/01/99\r\n'              # exp. date
                u'{self.usernum}\r\n'
                u'{self.xferprotocol}\r\n'
                u'0\r\n'                     # total num. uploads
                u'0\r\n'                     # total num, downloads
                u'0\r\n'                     # daily d/l limit
                u'9999999\r\n'               # return val/write val
                u'01/01/2001\r\n'            # birthdate
                # TODO
                u'C:\\XXX\r\n'               # filepaths to bbs files ...
                u'C:\\XXX\r\n'               # filepaths to bbs files ...
                u'{self.sysopname}\r\n'      # sysop's name
                u'{self.alias}\r\n'          # user's alias
                u'00:05\r\n'                 # event time(?)
                u'Y\r\n'                     # error-correcting connection
                u'Y\r\n'                     # is ANSI in NG mode?
                u'Y\r\n'                     # use record locking?
                u'7\r\n'                     # default color ..
                u'{self.remaining_mins}\r\n'
                u'09/09/99\r\n'              # last new file scan,
                u'{self.lastcall_time}\r\n'  # time of this call
                u'{self.lastcall_time}\r\n'  # time of last call
                u'9999\r\n'                  # max daily files
                u'0\r\n'                     # num. files today
                u'0\r\n'                     # u/l Kb today
                u'0\r\n'                     # d/l Kb today
                u'None\r\n'                  # user comment
                u'0\r\n'                     # doors opened
                u'0\n'                       # msgs left
                .format(self=self))

    def _get_door32(self):
        """ Return door32.sys-formatted dropfile content. """
        return (u'{self.comtype}\r\n'
                u'{self.comhandle}\r\n'
                u'{self.comspeed}\r\n'
                u'{self.systemname}\r\n'
                u'{self.usernum}\r\n'
                u'{self.fullname}\r\n'
                u'{self.alias}\r\n'
                u'{self.securitylevel}\r\n'
                u'{self.remaining_mins}\r\n'
                u'1\r\n'                  # emulation (1=ansi)
                u'{self.node}\n'
                .format(self=self))

    def _get_callinfo(self):
        """ Return callinfo.BBS-formatted dropfile content. """
        return (u'{self.alias}\r\n'
                u'{self.comspeed}\r\n'
                u'{self.location}\r\n'
                u'{self.securitylevel}\r\n'
                u'{self.remaining_mins}\r\n'
                u'COLOR\r\n'              # COLOR=ansi
                u'{self.password}\r\n'
                u'{self.usernum}\r\n'
                u'{self.time_used}\r\n'
                u'01:23\r\n'              # 1
                u'01:23 01/02/90\r\n'     # ..
                u'ABCDEFGH\r\n'           # ..
                u'0\r\n'                  # ..
                u'99\r\n'                 # ..
                u'0\r\n'                  # ..
                u'9999\r\n'               # 7 unknown fields,
                u'123-456-7890\r\n'       # phone number
                u'01/01/90 02:34\r\n'     # unknown date/time
                u'NOVICE\r\n'             # expert mode (off)
                u'{self.xferprotocol}\r\n'
                u'01/01/90\r\n'           # unknown date
                u'{self.numcalls}\r\n'
                u'{self.pageheight}\r\n'
                u'0\r\n'                  # ptr to new msgs?
                u'0\r\n'                  # total u/l
                u'0\r\n'                  # total d/l
                u'{self.parity}\r\n'      # ?? like 8,N,1 ??
                u'REMOTE\r\n'             # local or remote?
                u'{self.comport}\r\n'
                u'{self.comspeed}\r\n'
                u'FALSE\r\n'              # unknown,
                u'Normal Connection\r\n'  # unknown,
                u'01/02/94 01:20\r\n'     # unknown date/time
                u'0\r\n'                  # task #
                u'1\n'                    # door #
                .format(self=self))

    def _get_dorinfo(self):
        """ Return DORINFO.DEF-formatted dropfile content. """
        return (u'{self.systemname}\r\n'
                u'{self.sysopname}\r\n'   # sysop f.name
                u'{self.sysopname}\r\n'   # sysop l.name
                u'{self.comport}\r\n'
                u'{self.comspeed}\r\n'
                u'0\r\n'                  # "networked"?
                u'{self.alias}\r\n'       # user f.name
                u'{self.alias}\r\n'       # user l.name
                u'{self.location}\r\n'
                u'1\r\n'                  # term (1=ansi)
                u'{self.securitylevel}\r\n'
                u'{self.remaining_mins}\r\n'
                u'-1\n'                   # fossil (-1=external)
                .format(self=self))


class Door(object):

    """ Spawns a subprocess and pipes input and output over bbs session. """

    time_ipoll = 0.05
    time_opoll = 0.05
    blocksize = 7680
    master_fd = None

    def __init__(self, cmd='/bin/uname', args=(), env=None, cp437=False):
        """
        Class initializer.

        :param str cmd: full path of command to execute.
        :param tuple args: command arguments as tuple.
        :param bool cp437: When true, forces decoding of external program as
                           codepage 437.  This is the most common encoding used
                           by DOS doors.
        :param dict env: Environment variables to extend to the sub-process.
                         You should more than likely specify values for TERM,
                         PATH, HOME, and LANG.
        """
        self._session, self._term = getsession(), getterminal()
        self.cmd = cmd
        if isinstance(args, tuple):
            self.args = (self.cmd,) + args
        elif isinstance(args, list):
            self.args = [self.cmd, ] + args
        else:
            raise ValueError('args must be tuple or list')

        self.log = logging.getLogger(__name__)
        self.env = (env or {}).copy()
        self.env.update(
            {'LANG': env.get('LANG', 'en_US.UTF-8'),
             'TERM': env.get('TERM', self._term.kind),
             'PATH': env.get('PATH', get_ini('door', 'path')),
             'HOME': env.get('HOME', os.getenv('HOME')),
             'LINES': str(self._term.height),
             'COLUMNS': str(self._term.width),
             })

        self.cp437 = cp437
        self._utf8_decoder = codecs.getincrementaldecoder('utf8')()

    def run(self):
        """
        Begin door execution.

        pty.fork() is called, child process calls execvpe() while the parent
        process pipes session IPC data to and from the slave pty, until the
        child process exits.
        """
        try:
            import termios
            import fcntl
            import pty
        except ImportError as err:
            raise OSError('door support not (yet) supported on {0} platform.'
                          .format(sys.platform.lower()))

        self.log.debug('os.execvpe(cmd={self.cmd}, args={self.args}, '
                       'env={self.env})'.format(self=self))
        try:
            pid, self.master_fd = pty.fork()
        except OSError as err:
            # too many open files, out of memory, no such file/directory
            self.log.error('OSError in pty.fork(): %s', err)
            return

        # child process
        if pid == pty.CHILD:
            sys.stdout.flush()

            # send initial screen size
            _bytes = struct.pack(
                'HHHH', self._term.height, self._term.width, 0, 0)

            # pylint: disable=E1101
            #         Instance of 'DummyStream' has no 'fileno' member
            fcntl.ioctl(sys.stdout.fileno(), termios.TIOCSWINSZ, _bytes)

            try:
                os.execvpe(self.cmd, self.args, self.env)

            except OSError as err:
                # we cannot log an exception, only print to stderr and have
                # it captured by the parent process; this is because our
                # 'log' instance is dangerously forked, and any attempt to
                # communicate with multiprocessing pipes, loggers, etc. will
                # cause the value and state of many various file descriptors
                # to become corrupted, as our file descriptors are shared.

                sys.stderr.write('%s\n' % (err,))

            # pylint: disable=W0212
            #         Access to a protected member _exit of a client class
            os._exit(1)

        # parent process
        #
        # execute self._loop() and catch all i/o and o/s errors
        try:
            self.log.info('exec/%s: %r, env=%r', pid, self.args, self.env)
            self._loop()

        except IOError as err:
            self.log.error('IOError: %s', err)

        except OSError as err:
            # errno 5 is OK: it occurs when a read() call occurs after
            # sys.stdout has been closed by the child.
            if err.errno != 5:
                self.log.error('OSError: %s', err)

        (pid, status) = os.waitpid(pid, 0)
        res = status >> 8

        log_func = self.log.error if res != 0 else self.log.debug
        log_func('%s child %s exit %d', self.cmd, pid, res)

        os.close(self.master_fd)
        return res

    def input_filter(self, data):
        """
        Derive and modify to implement a keyboard-input filter.

        When keyboard input is detected, this method may filter such input.
        This base class method simply returns data as-is.
        """
        return data

    def output_filter(self, data):
        """
        Filter output (performs cp437 encoding).

        Given door output in bytes, if 'cp437' is specified in class
        initializer, convert to utf8 glyphs using cp437 encoding;
        otherwise decode output naturally as utf8.
        """
        if self.cp437:
            # cp437 bytes don't need to be incrementally decoded, each
            # byte is always final.
            return data.decode('cp437_art')

        # utf-8, however, may be read mid-stream of a multibyte sequence.
        decoded = list()
        for byte in data:
            ucs = self._utf8_decoder.decode(byte, final=False)
            if ucs is not None:
                decoded.append(ucs)
        return u''.join(decoded)

    def resize(self):
        """ Signal resize of terminal to pty. """
        import termios
        import fcntl
        _bytes = struct.pack('HHHH',
                             self._term.height,
                             self._term.width,
                             0, 0)
        fcntl.ioctl(self.master_fd, termios.TIOCSWINSZ, _bytes)

    def _loop(self):
        """ Main event loop, polling i/o of pty and session. """
        while True:
            # block up to self.time_opoll for screen output
            if self.master_fd == -1:
                # pty file descriptor closed by child,
                # early termination!
                break

            rlist = (self.master_fd,)
            ret_tuple = select.select(rlist, (), (), self.time_opoll)
            if self.master_fd in ret_tuple[0]:
                data = os.read(self.master_fd, self.blocksize)
                if 0 == len(data):
                    break
                echo(self.output_filter(data))

            # block up to self.time_ipoll for keyboard input
            event, data = self._session.read_events(
                ('refresh', 'input',), self.time_ipoll)

            if event == 'refresh' and data[0] == 'resize':
                self.resize()

            elif event == 'input':
                data = self.input_filter(data)
                if 0 != len(data):
                    n_written = os.write(self.master_fd, data)
                    if n_written != len(data):
                        # we wrote none or some of our keyboard input, but
                        # not all. re-buffer remaining bytes back into
                        # session for next poll
                        self._session.buffer_input(data[n_written:])
                        self.log.warn('re-buffer_input(%r)!', data[n_written:])


class DOSDoor(Door):

    """
    Door-derived class with special handlers for executing dosemu.

    This Door-derived class removes the "report cursor position" query
    sequence, which is sent by DOSEMU on startup. It also removes the "switch
    to alternate screen mode" set and reset (blessings terminals provide this
    with the context manager, using statement ``with term.fullscreen():``).

    It would appear that any early keyboard input received (esp. in response
    to "report cursor position") prior to DOOR execution in DOSEMU causes all
    input to be bitshifted and invalid and/or broken.

    This class resolves that issue by overriding ``output_filter`` to remove
    such sequences, and ``input_filter`` which only allows input after a few
    seconds have elapsed.
    """

    #: regular expression of sequences to be replaced by ``term.clear``
    #: during ``START_BLOCK`` delay in ``output_filter``
    RE_REPWITH_CLEAR = (r'\033\[('
                        r'1;80H.*\033\[1;1H'
                        r'|H\033\[2J'
                        r'|\d+;1H.*\033\[1;1H'
                        r')')

    #: regular expression of sequences to strip entirely during
    #: ``START_BLOCK`` delay in ``output_filter``.
    RE_REPWITH_NONE = (r'\033\[('
                       r'6n'
                       r'|\?1049[lh]'
                       r'|\d+;\d+r'
                       r'|1;1H\033\[\dM)')

    #: Number of seconds to allow to elapse for ``input_filter`` and
    #: ``output_filter`` as a workaround for stripping startup sequences
    #: and working around a strange keyboard input bug.
    START_BLOCK = 4.0

    def __init__(self, cmd='/bin/uname', args=(), env=None, cp437=True):
        """
        Class initializer.

        :param str cmd: full path of command to execute.
        :param tuple args: command arguments as tuple.
        :param bool cp437: When true, forces decoding of external program as
                           codepage 437.  This is the most common encoding used
                           by DOS doors.
        :param dict env: Environment variables to extend to the sub-process.
                         You should more than likely specify values for TERM,
                         PATH, HOME, and LANG.
        """
        Door.__init__(self, cmd=cmd, args=args, env=env, cp437=cp437)
        self._re_trim_clear = re.compile(self.RE_REPWITH_CLEAR,
                                         flags=re.DOTALL)
        self._re_trim_none = re.compile(self.RE_REPWITH_NONE,
                                        flags=re.DOTALL)
        self._replace_clear = u''.join((self._term.move(25, 0),
                                        (u'\r\n' * 25),
                                        self._term.home))

    def output_filter(self, data):
        """ filter screen output (removes dosemu startup sequences). """
        data = Door.output_filter(self, data)
        if self._stime is not None and (
                time.time() - self._stime < self.START_BLOCK):
            data = re.sub(pattern=self._re_trim_clear,
                          repl=(self._replace_clear), string=data)
            data = re.sub(pattern=self._re_trim_none,
                          repl=u'\r\n', string=data)
        return data

    def input_filter(self, data):
        """ filter keyboard input (used for "throway" bug workaround). """
        return data if time.time() - self._stime > self.START_BLOCK else u''

    def resize(self):
        """ Signal resize of terminal to DOS -- does nothing. """
        pass

    def run(self):
        """
        Begin door execution.

        pty.fork() is called, child process calls execvpe() while the parent
        process pipes telnet session IPC data to and from the slave pty until
        child process exits.

        On exit, DOSDoor flushes any keyboard input; DOSEMU appears to send
        various terminal reset sequences that may cause a reply to be received
        on input, and later as an invalid menu command.
        """
        self._stime = time.time()

        Door.run(self)

        # fight against 'set scrolling region' by resetting, LORD
        # contains, for example: \x1b[3;22r after 'E'nter the realm
        echo(u''.join((self._term.normal,
                       self._term.move(self._term.height, self._term.width),
                       u"\x1b[r",
                       self._term.move(self._term.height, 0),
                       u'\r\n\r\n')))

        # flush any previously decoded but unreceived keystrokes,
        # and any unprocessed input from telnet session not yet processed.
        self._term.kbflush()
        self._session.flush_event('input')
